# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rucio.common.config import config_get
from rucio.common.types import InternalScope
from rucio.core.account import add_account_attribute
from rucio.core.scope import add_scope
from rucio.db.sqla.constants import DatabaseOperationType
from rucio.db.sqla.session import db_session
from rucio.gateway.permission import has_permission
from rucio.tests.common import scope_name_generator, skip_non_belleii


class TestPermissionCoreGateway:
    """
    Test the Permission Core and Gateway
    """

    usr = 'jdoe'

    def test_permission_add_did(self, vo, root_account):
        """ PERMISSION(CORE): Check permission to add a DID"""
        scope = scope_name_generator()
        with db_session(DatabaseOperationType.WRITE) as session:
            add_scope(scope=InternalScope(scope, vo=vo), account=root_account, session=session)
        assert has_permission(issuer='panda', action='add_did', kwargs={'scope': scope}, vo=vo)
        assert not has_permission(issuer='spock', action='add_did', kwargs={'scope': scope}, vo=vo)

    def test_permission_add_account(self, vo):
        """ PERMISSION(CORE): Check permission to add account """
        assert has_permission(issuer='root', action='add_account', kwargs={'account': 'account1'}, vo=vo)
        assert not has_permission(issuer='self.usr', action='add_account', kwargs={'account': 'account1'}, vo=vo)

    def test_permission_add_scope(self, vo, random_account):
        """ PERMISSION(CORE): Check permission to add scope """
        assert has_permission(issuer='root', action='add_scope', kwargs={'account': 'root'}, vo=vo)
        assert not has_permission(issuer=random_account.external, action='add_scope', kwargs={'account': random_account.external}, vo=vo)
        add_account_attribute(random_account, 'admin', True)
        assert has_permission(issuer=random_account.external, action='add_scope', kwargs={'account': random_account.external}, vo=vo)

    @skip_non_belleii
    def test_permission_add_scope_admin(self, vo, random_account):
        """ PERMISSION(CORE): Check permission to add scope with scope_admin attribute (Belle II)"""
        assert not has_permission(issuer=random_account.external, action='add_scope', kwargs={'account': random_account.external}, vo=vo)
        add_account_attribute(random_account, 'scope_admin', True)
        assert has_permission(issuer=random_account.external, action='add_scope', kwargs={'account': random_account.external}, vo=vo)

    def test_permission_get_auth_token_user_pass(self, vo):
        """ PERMISSION(CORE): Check permission to get_auth_token_user_pass """
        assert has_permission(issuer='root', action='get_auth_token_user_pass', kwargs={'account': 'root', 'username': 'ddmlab', 'password': 'secret'}, vo=vo)
        assert not has_permission(issuer='root', action='get_auth_token_user_pass', kwargs={'account': self.usr, 'username': 'ddmlab', 'password': 'secret'}, vo=vo)

    def test_permission_get_auth_token_x509(self, vo):
        """ PERMISSION(CORE): Check permission to get_auth_token_x509 """
        dn = config_get('bootstrap', 'x509_identity')
        assert has_permission(issuer='root', action='get_auth_token_x509', kwargs={'account': 'root', 'dn': dn}, vo=vo)
        assert not has_permission(issuer='root', action='get_auth_token_x509', kwargs={'account': self.usr, 'dn': dn}, vo=vo)

    def test_permission_get_auth_token_gss(self, vo):
        """ PERMISSION(CORE): Check permission to get_auth_token_gss """
        gsscred = 'rucio-dev@CERN.CH'
        assert has_permission(issuer='root', action='get_auth_token_gss', kwargs={'account': 'root', 'gsscred': gsscred}, vo=vo)
        assert not has_permission(issuer='root', action='get_auth_token_gss', kwargs={'account': self.usr, 'gsscred': gsscred}, vo=vo)

    def test_permission_update_rule_boost(self, vo):
        kwargs = {'options': {'boost_rule': True}}
        assert has_permission(issuer='root', action='update_rule', kwargs=kwargs, vo=vo)
        assert not has_permission(issuer='jdoe', action='update_rule', kwargs=kwargs, vo=vo)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import remove
from os.path import basename

import pytest

from rucio.common.utils import execute
from rucio.common.utils import generate_uuid as uuid

MARKER = '$ > '


def delete_rules(did):
    # get the rules for the file
    print('Deleting rules')
    cmd = "rucio rule list {0} | grep {0} | cut -f1 -d\\ ".format(did)
    print(cmd)
    exitcode, out, err = execute(cmd)
    print(out, err)
    rules = out.split()
    # delete the rules for the file
    for rule in rules:
        cmd = "rucio rule remove {0}".format(rule)
        print(cmd)
        exitcode, out, err = execute(cmd)


@pytest.mark.noparallel(reason='uses pre-defined RSE')
class TestRucioServer:

    def test_ping(self):
        """CLIENT (USER): rucio ping"""
        cmd = 'rucio ping'
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(out, err)
        assert exitcode == 0

    def test_whoami(self):
        """CLIENT (USER): rucio whoami"""
        cmd = 'rucio whoami'
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(out, err)
        assert exitcode == 0

    def test_upload_download(self, file_factory, scope_and_rse):
        """CLIENT(USER): rucio upload files to dataset/download dataset"""

        scope, rse = scope_and_rse

        if rse is None:
            return

        tmp_file1 = file_factory.file_generator()
        tmp_file2 = file_factory.file_generator()
        tmp_file3 = file_factory.file_generator()
        tmp_dsn = 'tests.rucio_client_test_server_' + uuid()

        # Adding files to a new dataset
        cmd = 'rucio upload --rse {0} --scope {1} {2} {3} {4} {1}:{5}'.format(rse, scope, tmp_file1, tmp_file2, tmp_file3, tmp_dsn)
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(out)
        print(err)
        remove(tmp_file1)
        remove(tmp_file2)
        remove(tmp_file3)
        assert exitcode == 0

        # List the files
        cmd = 'rucio did content list {0}:{1}'.format(scope, tmp_dsn)
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(out)
        print(err)
        assert exitcode == 0

        # List the replicas
        cmd = 'rucio replica list file {0}:{1}'.format(scope, tmp_dsn)
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(out)
        print(err)
        assert exitcode == 0

        # Downloading dataset
        cmd = 'rucio download --dir /tmp/ {0}:{1}'.format(scope, tmp_dsn)
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(out)
        print(err)
        # The files should be there
        cmd = 'ls /tmp/{0}/*'.format(tmp_dsn)
        print(MARKER + cmd)
        exitcode, out, err = execute(cmd)
        print(err, out)
        assert exitcode == 0
        assert tmp_file1.name in out
        assert tmp_file2.name in out
        assert tmp_file3.name in out

        # cleaning
        remove('/tmp/{0}/'.format(tmp_dsn) + basename(tmp_file1))
        remove('/tmp/{0}/'.format(tmp_dsn) + basename(tmp_file2))
        remove('/tmp/{0}/'.format(tmp_dsn) + basename(tmp_file3))
        added_dids = ['{0}:{1}'.format(scope, did) for did in (basename(tmp_file1), basename(tmp_file2), basename(tmp_file3), tmp_dsn)]

        # Delete rules
        for did in added_dids:
            delete_rules(did)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import pytest

import rucio.db.sqla.util
from rucio.common import exception
from rucio.daemons.abacus import account, collection_replica, rse
from rucio.daemons.atropos import atropos
from rucio.daemons.automatix import automatix
from rucio.daemons.badreplicas import minos, minos_temporary_expiration, necromancer
from rucio.daemons.cache import consumer
from rucio.daemons.conveyor import finisher, poller, preparer, receiver, stager, submitter, throttler
from rucio.daemons.follower import follower
from rucio.daemons.hermes import hermes
from rucio.daemons.judge import cleaner, evaluator, injector, repairer
from rucio.daemons.oauthmanager import oauthmanager
from rucio.daemons.reaper import dark_reaper, reaper
from rucio.daemons.replicarecoverer import suspicious_replica_recoverer
from rucio.daemons.tracer import kronos
from rucio.daemons.transmogrifier import transmogrifier
from rucio.daemons.undertaker import undertaker

DAEMONS = [
    account,
    collection_replica,
    rse,
    atropos,
    automatix,
    minos,
    minos_temporary_expiration,
    necromancer,
    consumer,
    finisher,
    poller,
    receiver,
    stager,
    submitter,
    throttler,
    preparer,
    follower,
    hermes,
    cleaner,
    evaluator,
    injector,
    repairer,
    oauthmanager,
    dark_reaper,
    reaper,
    suspicious_replica_recoverer,
    kronos,
    transmogrifier,
    undertaker,
]

ids = [mod.__name__ for mod in DAEMONS]


@pytest.mark.parametrize('daemon', argvalues=DAEMONS, ids=ids)
@mock.patch('rucio.db.sqla.util.is_old_db')
def test_fail_on_old_database(mock_is_old_db, daemon):
    """ DAEMON: Test daemon failure on old database """
    mock_is_old_db.return_value = True
    assert rucio.db.sqla.util.is_old_db() is True

    with pytest.raises(exception.DatabaseException, match='Database was not updated, daemon won\'t start'):
        daemon.run()

    assert mock_is_old_db.call_count > 1



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import shutil
from random import choice
from string import ascii_uppercase
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from rucio.client.client import Client
from rucio.client.uploadclient import UploadClient
from rucio.common.checksum import adler32, md5
from rucio.common.config import config_add_section, config_set
from rucio.common.constants import RseAttr
from rucio.common.exception import InputValidationError, NoFilesUploaded, NotAllFilesUploaded, ResourceTemporaryUnavailable
from rucio.common.types import InternalScope
from rucio.common.utils import execute, generate_uuid
from rucio.core.rse import add_protocol, add_rse_attribute
from rucio.tests.common import did_name_generator

if TYPE_CHECKING:
    from rucio.common.types import FileToUploadDict


@pytest.fixture
def upload_client():
    logger = logging.getLogger('upload_client')
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)
    return UploadClient(logger=logger)


@pytest.fixture
def rse(containerized_rses, rse_factory):
    if len(containerized_rses) > 0:
        rse, _ = containerized_rses[0]
    else:
        rse, _ = rse_factory.make_posix_rse()
    return rse


@pytest.fixture
def scope(vo, containerized_rses, test_scope, mock_scope):
    if len(containerized_rses) > 0:
        return str(test_scope)
    else:
        return str(mock_scope)


@pytest.mark.parametrize("file_config_mock", [
    {},  # Use rucio.cfg as-is.
    pytest.param(
        {  # Remove "account" from the "[client]" section.
            "removes": [('client', 'account')]
        }, marks=pytest.mark.skipif('SUITE' in os.environ and os.environ['SUITE'] == 'multi_vo', reason="See https://github.com/rucio/rucio/issues/7394"))
    ], indirect=True)
def test_upload_single(file_config_mock, rse, scope, upload_client, download_client, file_factory):
    local_file = file_factory.file_generator()
    download_dir = file_factory.base_dir
    fn = os.path.basename(local_file)

    item: FileToUploadDict = {
        'path': local_file,
        'rse': rse,
        'did_scope': scope,
        'did_name': fn,
        'guid': generate_uuid()
    }

    # upload a file
    status = upload_client.upload(items=[item])
    assert status == 0

    # download the file
    did = f"{scope}:{fn}"
    download_client.download_dids([{'did': did, 'base_dir': download_dir}])

    # match checksums
    downloaded_file = f"{download_dir}/{scope}/{fn}"
    assert adler32(local_file) == adler32(downloaded_file)


def test_upload_multi(rse, scope, upload_client, download_client, file_factory):
    local_file1 = file_factory.file_generator(use_basedir=True)
    local_file2 = file_factory.file_generator(use_basedir=True)
    download_dir = file_factory.base_dir

    fn1 = os.path.basename(local_file1)
    fn2 = os.path.basename(local_file2)

    items: list[FileToUploadDict] = [
        {
            'path': local_file1,
            'rse': rse,
            'did_scope': scope,
            'did_name': fn1,
            'guid': generate_uuid(),
        },
        {
            'path': local_file2,
            'rse': rse,
            'did_scope': scope,
            'did_name': fn2,
            'guid': generate_uuid(),
        },
    ]

    status = upload_client.upload(items=items)
    assert status == 0
    # download the files
    did1 = f"{scope}:{fn1}"
    did2 = f"{scope}:{fn2}"
    download_client.download_dids([
        {'did': did1, 'base_dir': download_dir},
        {'did': did2, 'base_dir': download_dir}
    ])

    # match checksums
    downloaded_file1 = f"{download_dir}/{scope}/{fn1}"
    assert adler32(local_file1) == adler32(downloaded_file1)

    downloaded_file2 = f"{download_dir}/{scope}/{fn2}"
    assert adler32(local_file2) == adler32(downloaded_file2)


def test_upload_file_already_exists_single(rse, scope, upload_client, file_factory):
    traces = []
    local_file = file_factory.file_generator()

    item: FileToUploadDict = {
        'path': local_file,
        'rse': rse,
        'did_scope': scope,
    }

    # upload the file
    upload_client.upload(items=[item])
    # re_upload the file
    with pytest.raises(NoFilesUploaded):
        upload_client.upload(items=[item], traces_copy_out=traces)
    assert len(traces) == 1 and traces[0]['stateReason'] == 'File already exists'


def test_upload_file_already_exists_multi(rse, scope, upload_client, file_factory):
    traces = []
    local_file = file_factory.file_generator()

    items: list[FileToUploadDict] = [
        {
            'path': local_file,
            'rse': rse,
            'did_scope': scope,
        },
        {
            'path': local_file,
            'rse': rse,
            'did_scope': scope,
        }
    ]

    # upload the file twice in the same upload command
    with pytest.raises(NotAllFilesUploaded):
        upload_client.upload(items=items, traces_copy_out=traces)

    assert len(traces) == 2 and traces[1]['stateReason'] == 'File already exists'


def test_upload_source_not_found(rse, scope, upload_client):
    item: FileToUploadDict = {
        'path': 'non_existent_local_file',
        'rse': rse,
        'did_scope': scope,
    }

    with pytest.raises(InputValidationError):
        upload_client.upload(items=[item])


def test_multiple_protocols_same_scheme(rse_factory, upload_client, mock_scope, file_factory):
    """ Upload (CLIENT): Ensure domain correctly selected when multiple protocols exist with the same scheme """

    rse, rse_id = rse_factory.make_rse()

    # Ensure client site and rse site are identical. So that "lan" is preferred.
    add_rse_attribute(rse_id, RseAttr.SITE, 'ROAMING')

    add_protocol(rse_id, {'scheme': 'file',
                          'hostname': 'file-wan.aperture.com',
                          'port': 0,
                          'prefix': '/prefix1/',
                          'impl': 'rucio.rse.protocols.posix.Default',
                          'domains': {
                              'lan': {'read': None, 'write': None, 'delete': None},
                              'wan': {'read': 1, 'write': 1, 'delete': 1}}})
    add_protocol(rse_id, {'scheme': 'file',
                          'hostname': 'file-lan.aperture.com',
                          'port': 0,
                          'prefix': '/prefix2/',
                          'impl': 'rucio.rse.protocols.posix.Default',
                          'domains': {
                              'lan': {'read': 1, 'write': 1, 'delete': 1},
                              'wan': {'read': None, 'write': None, 'delete': None}}})
    add_protocol(rse_id, {'scheme': 'root',
                          'hostname': 'root.aperture.com',
                          'port': 1403,
                          'prefix': '/prefix3/',
                          'impl': 'rucio.rse.protocols.xrootd.Default',
                          'domains': {
                              'lan': {'read': 2, 'write': 2, 'delete': 2},
                              'wan': {'read': 2, 'write': 2, 'delete': 2}}})

    # Upload a file
    path = file_factory.file_generator()
    name = os.path.basename(path)

    item: FileToUploadDict = {
        'path': path,
        'rse': rse,
        'did_scope': str(mock_scope),
        'did_name': name,
        'guid': generate_uuid(),
    }
    summary_path = file_factory.base_dir / 'summary'
    upload_client.upload(items=[item], summary_file_path=summary_path)

    # Verify that the lan protocol was used for the upload
    with open(summary_path) as json_file:
        data = json.load(json_file)
        assert 'file-lan.aperture.com' in data['{}:{}'.format(mock_scope, name)]['pfn']


def test_upload_file_with_impl(rse_factory, upload_client, mock_scope, file_factory):
    """ Upload (CLIENT): Ensure the module associated to the impl value is called """

    impl = 'xrootd'
    rse_name, rse_id = rse_factory.make_rse()

    add_protocol(rse_id, {'scheme': 'file',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.posix.Default',
                          'domains': {
                              'lan': {'read': 1, 'write': 1, 'delete': 1},
                              'wan': {'read': 1, 'write': 1, 'delete': 1}}})
    add_protocol(rse_id, {'scheme': 'root',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.xrootd.Default',
                          'domains': {
                              'lan': {'read': 2, 'write': 2, 'delete': 2},
                              'wan': {'read': 2, 'write': 2, 'delete': 2}}})

    path = file_factory.file_generator()
    name = os.path.basename(path)

    item: FileToUploadDict = {
        'path': path,
        'rse': rse_name,
        'did_scope': str(mock_scope),
        'did_name': name,
        'guid': generate_uuid(),
        'impl': impl
    }

    with TemporaryDirectory() as tmp_dir:
        with patch('rucio.rse.protocols.%s.Default.put' % impl, side_effect=lambda pfn, dest, dir, **kw: shutil.copy(path, tmp_dir)) as mock_put, \
                patch('rucio.rse.protocols.%s.Default.connect' % impl), \
                patch('rucio.rse.protocols.%s.Default.exists' % impl, side_effect=lambda pfn, **kw: False), \
                patch('rucio.rse.protocols.%s.Default.delete' % impl), \
                patch('rucio.rse.protocols.%s.Default.rename' % impl), \
                patch('rucio.rse.protocols.%s.Default.stat' % impl, side_effect=lambda pfn: {'filesize': os.stat(path)[os.path.stat.ST_SIZE], 'adler32': adler32(path)}), \
                patch('rucio.rse.protocols.%s.Default.close' % impl):
            mock_put.__name__ = "mock_put"
            upload_client.upload(items=[item])
            mock_put.assert_called()


def test_upload_file_with_supported_protocol(rse_factory, upload_client, mock_scope, file_factory):
    """ Upload (CLIENT): Ensure the module associated to the first protocol supported by both the remote and local config is called """

    rse_name, rse_id = rse_factory.make_rse()

    # FIXME:
    # The correct order to test should actually be ssh,xrootd,posix
    # However the preferred_impl is not working correctly.
    # Once preferred_impl is fixed, this should be changed back
    add_protocol(rse_id, {'scheme': 'scp',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.posix.Default',
                          'domains': {
                              'lan': {'read': 1, 'write': 1, 'delete': 1},
                              'wan': {'read': 1, 'write': 1, 'delete': 1}}})
    add_protocol(rse_id, {'scheme': 'root',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.xrootd.Default',
                          'domains': {
                              'lan': {'read': 2, 'write': 2, 'delete': 2},
                              'wan': {'read': 2, 'write': 2, 'delete': 2}}})
    add_protocol(rse_id, {'scheme': 'file',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.ssh.Default',
                          'domains': {
                              'lan': {'read': 3, 'write': 3, 'delete': 3},
                              'wan': {'read': 3, 'write': 3, 'delete': 3}}})

    path = file_factory.file_generator()
    name = os.path.basename(path)

    item: FileToUploadDict = {
        'path': path,
        'rse': rse_name,
        'did_scope': str(mock_scope),
        'did_name': name,
        'guid': generate_uuid()
    }

    status = upload_client.upload(items=[item])
    assert status == 0


def test_upload_file_with_supported_protocol_from_config(rse_factory, upload_client, mock_scope, file_factory):
    """ Upload (CLIENT): Ensure the module associated to the first protocol supported by both the remote and local config read from rucio.cfg is called """

    rse_name, rse_id = rse_factory.make_rse()

    # FIXME:
    # The correct order to test should actually be ssh,xrootd,posix
    # However the preferred_impl is not working correctly.
    # Once preferred_impl is fixed, this should be changed back
    add_protocol(rse_id, {'scheme': 'scp',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.xrootd.Default',
                          'domains': {
                              'lan': {'read': 1, 'write': 1, 'delete': 1},
                              'wan': {'read': 1, 'write': 1, 'delete': 1}}})
    add_protocol(rse_id, {'scheme': 'file',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.posix.Default',
                          'domains': {
                              'lan': {'read': 2, 'write': 2, 'delete': 2},
                              'wan': {'read': 2, 'write': 2, 'delete': 2}}})
    add_protocol(rse_id, {'scheme': 'root',
                          'hostname': '%s.cern.ch' % rse_id,
                          'port': 0,
                          'prefix': '/test/',
                          'impl': 'rucio.rse.protocols.ssh.Default',
                          'domains': {
                              'lan': {'read': 3, 'write': 3, 'delete': 3},
                              'wan': {'read': 3, 'write': 3, 'delete': 3}}})

    config_add_section('upload')
    config_set('upload', 'preferred_impl', 'rclone, xrootd')

    supported_impl = 'xrootd'

    path = file_factory.file_generator()
    name = os.path.basename(path)

    item: FileToUploadDict = {
        'path': path,
        'rse': rse_name,
        'did_scope': str(mock_scope),
        'did_name': name,
        'guid': generate_uuid()
    }

    with TemporaryDirectory() as tmp_dir:
        with patch('rucio.rse.protocols.%s.Default.put' % supported_impl, side_effect=lambda pfn, dest, dir, **kw: shutil.copy(path, tmp_dir)) as mock_put, \
                patch('rucio.rse.protocols.%s.Default.connect' % supported_impl), \
                patch('rucio.rse.protocols.%s.Default.exists' % supported_impl, side_effect=lambda pfn, **kw: False), \
                patch('rucio.rse.protocols.%s.Default.delete' % supported_impl), \
                patch('rucio.rse.protocols.%s.Default.rename' % supported_impl), \
                patch('rucio.rse.protocols.%s.Default.stat' % supported_impl, side_effect=lambda pfn: {'filesize': os.stat(path)[os.path.stat.ST_SIZE], 'adler32': adler32(path)}), \
                patch('rucio.rse.protocols.%s.Default.close' % supported_impl):
            mock_put.__name__ = "mock_put"
            upload_client.upload(items=[item])
            mock_put.assert_called()


def test_upload_file_ignore_availability(rse_factory, scope, upload_client, file_factory, rucio_client):
    rse, rse_id = rse_factory.make_posix_rse()
    rucio_client.update_rse(rse, {'availability_write': False})
    local_file = file_factory.file_generator()

    item: FileToUploadDict = {
        'path': local_file,
        'rse': rse,
        'did_scope': scope,
    }

    status = upload_client.upload(items=[item], ignore_availability=True)
    assert status == 0


@pytest.fixture
def upload_client_registration_fail():
    logger = logging.getLogger('upload_client')
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)

    # modify the client object used by upload_client so that replica registration fails
    class RegistrationFailureClient(Client):

        def __init__(self, **args):
            super(RegistrationFailureClient, self).__init__(**args)

        def update_replicas_states(self, rse, files):
            # simulate server timing out
            raise ResourceTemporaryUnavailable

    return UploadClient(logger=logger, _client=RegistrationFailureClient())


def test_upload_registration_fail(rse, scope, upload_client_registration_fail, file_factory):
    local_file = file_factory.file_generator()
    fn = os.path.basename(local_file)

    # upload a file and check that exception is raised
    with pytest.raises(NoFilesUploaded):
        item: FileToUploadDict = {
            'path': local_file,
            'rse': rse,
            'did_scope': scope,
            'did_name': fn,
            'guid': generate_uuid()
        }

        upload_client_registration_fail.upload(items=[item])


@pytest.mark.skipif('SUITE' in os.environ and os.environ['SUITE'] == 'client', reason="Requires DB access")
def test_upload_file_register_after_upload(rse, scope, upload_client, rucio_client, file_factory, vo):
    """CLIENT(USER): Rucio upload files with registration after upload"""
    local_file1 = file_factory.file_generator()
    local_file2 = file_factory.file_generator()
    local_file3 = file_factory.file_generator()

    local_file1_name = os.path.basename(local_file1)
    local_file2_name = os.path.basename(local_file2)
    local_file3_name = os.path.basename(local_file3)

    status = upload_client.upload([
        {
            "did_scope": scope,
            "rse": rse,
            "did_name": fn,
            "path": path,
            "register_after_upload": True
        }
        for fn, path in zip([local_file1_name, local_file2_name, local_file3_name], [local_file1, local_file2, local_file3])
    ])
    assert status == 0

    # Trying to upload again produces an error
    with pytest.raises(NoFilesUploaded):
        upload_client.upload([
            {
                "did_scope": scope,
                "rse": rse,
                "did_name": local_file1_name,
                "path": local_file1,
            }
        ])

    # removing replica -> file on RSE should be overwritten
    # (simulating an upload error, where a part of the file is uploaded but the replica is not registered)
    from sqlalchemy import and_, delete

    from rucio.db.sqla import models, session

    db_session = session.get_session()
    internal_scope = InternalScope(scope, vo=vo)
    for model in [models.RSEFileAssociation, models.ReplicaLock, models.ReplicationRule, models.DidMeta, models.DataIdentifier]:
        stmt = delete(
            model
        ).where(
            and_(
                model.name == local_file1_name,
                model.scope == internal_scope)
            )
        db_session.execute(stmt)
    db_session.commit()

    # After removing the replica, you can upload with the replica name however you like
    local_file4 = file_factory.file_generator()
    upload_client.upload([
        {"did_scope": scope, "rse": rse, "did_name": local_file1_name, "path": local_file4, "register_after_upload": True}])
    assert md5(local_file4) == [replica for replica in rucio_client.list_replicas(dids=[{'name': local_file1_name, 'scope': scope}])][0]['md5']


@pytest.mark.xfail(reason="Permission error needs to be resolved")
def test_upload_repeat_after_deletion(file_factory, rse_factory, upload_client, scope, rucio_client):
    """CLIENT(USER): Rucio upload repeated files"""
    # One of the files to upload is already catalogued but was removed
    rse, _ = rse_factory.make_posix_rse()

    local_file1 = file_factory.file_generator()
    local_file1_name = os.path.basename(local_file1)

    upload_client.upload([
        {
            "did_scope": scope,
            "rse": rse,
            "did_name": local_file1_name,
            "path": local_file1,
        }
    ])

    # Find and delete the rule
    print([f for f in rucio_client.list_replication_rules({'name': local_file1_name})])
    rule_id = [f['id'] for f in rucio_client.list_replication_rules({'name': local_file1_name})][0]
    rucio_client.delete_replication_rule(rule_id)

    # Delete the file from the RSE
    # ??? Why does root not have permission to delete the file?
    rucio_client.delete_replicas(rse, [{'scope': scope, 'name': local_file1_name}])

    # File can be re-uploaded
    status = upload_client.upload([
        {
            "did_scope": scope,
            "rse": rse,
            "did_name": local_file1_name,
            "path": local_file1,
        }
    ])
    assert status == 0


def test_upload_repeated_file_dataset(file_factory, rse_factory, rucio_client, upload_client, scope):
    """CLIENT(USER): Rucio upload repeated files to dataset"""
    # One of the files to upload is already in the dataset

    rse, _ = rse_factory.make_posix_rse()
    tmp_file1 = file_factory.file_generator()
    tmp_file2 = file_factory.file_generator()

    tmp_file1_name = os.path.basename(tmp_file1)
    tmp_file2_name = os.path.basename(tmp_file2)

    tmp_dataset = f"DSet{generate_uuid()}"
    # User uploads a dataset with a single file
    status = upload_client.upload([
        {"dataset_name": tmp_dataset, "dataset_scope": scope, "path": tmp_file1, "rse": rse, "did_scope": scope, "did_name": tmp_file1_name}
    ])
    assert status == 0

    # Other files are added to the same dataset
    status = upload_client.upload([
        {"dataset_name": tmp_dataset, "dataset_scope": scope, "path": tmp_file2, "rse": rse, "did_scope": scope, "did_name": tmp_file2_name}
    ])
    assert status == 0
    # All files are added to the same dataset
    files = [f['name'] for f in rucio_client.list_files(scope=scope, name=tmp_dataset)]
    assert tmp_file1_name in files
    assert tmp_file2_name in files


def test_upload_adds_md5digest(file_factory, rse_factory, rucio_client, upload_client, scope):
    """CLIENT(USER): Upload Checksums"""
    # user has a file to upload
    rse, _ = rse_factory.make_posix_rse()
    tmp_file = file_factory.file_generator()
    tmp_file_name = os.path.basename(tmp_file)
    file_md5 = md5(tmp_file)
    # user uploads file
    upload_client.upload([
        {
            "did_scope": scope,
            "rse": rse,
            "did_name": tmp_file_name,
            "path": tmp_file,
        }
    ])

    meta = rucio_client.get_metadata(scope=scope, name=tmp_file_name)
    assert 'md5' in meta
    assert meta['md5'] == file_md5


def test_upload_recursive_preserve_structure(rse_factory, rucio_client, upload_client, scope):
    """CLIENT(USER): Upload and preserve folder structure"""
    rse, _ = rse_factory.make_posix_rse()
    with TemporaryDirectory() as tmp_dir:
        folder1 = '%s/folder_%s' % (tmp_dir, generate_uuid())
        folder2 = '%s/folder_%s' % (tmp_dir, generate_uuid())
        folder3 = '%s/folder_%s' % (tmp_dir, generate_uuid())
        folder11 = '%s/folder_%s' % (folder1, generate_uuid())
        folder12 = '%s/folder_%s' % (folder1, generate_uuid())
        folder13 = '%s/folder_%s' % (folder1, generate_uuid())
        file1 = 'file_%s' % generate_uuid()
        file2 = 'file_%s' % generate_uuid()

        os.makedirs(folder1)
        os.makedirs(folder2)
        os.makedirs(folder3)
        os.makedirs(folder11)
        os.makedirs(folder12)
        os.makedirs(folder13)

        with open('%s/%s' % (folder11, file1), 'w') as f:
            f.write(generate_uuid())
        with open('%s/%s' % (folder2, file2), 'w') as f:
            f.write(generate_uuid())

        status = upload_client.upload([
            {
                "did_scope": scope,
                "path": tmp_dir,
                "recursive": True,
                "rse": rse,
            }
        ])
        assert status == 0

        contents = [f['name'] for f in rucio_client.list_content(scope=scope, name=tmp_dir.split('/')[-1])]
        assert len(contents) == 2
        assert folder1.split('/')[-1] in contents
        assert folder2.split('/')[-1] in contents
        assert folder3.split('/')[-1] not in contents  # Folder 3 is empty, and is skipped during upload

        contents = [f['name'] for f in rucio_client.list_content(scope=scope, name=folder1.split('/')[-1])]
        assert len(contents) == 1
        assert folder11.split('/')[-1] in contents
        assert folder12.split('/')[-1] not in contents  # 12 and 13 are empty, and are skipped during upload
        assert folder13.split('/')[-1] not in contents

        contents = [f['name'] for f in rucio_client.list_content(scope=scope, name=folder11.split('/')[-1])]
        assert len(contents) == 1
        assert file1 in contents

        contents = [f['name'] for f in rucio_client.list_content(scope=scope, name=folder2.split('/')[-1])]
        assert len(contents) == 1
        assert file2 in contents


@pytest.mark.parametrize("structure,is_valid_container", [
    ({"bad_container": ["file1", "file2", {"dataset": ["file3", "file4"]}]}, False),  # Cannot have both files and datasets in a container
    ({
        "valid_container": {
            "dataset": ["file1", "file2"]
            },
        "bad_container": {
            "dataset": ["file1", {"bad_dataset": ["file1"]}]
        }
    }, False),  # One valid container, another invalid container
    ({
        "container": {
            "bad_container": {
                "dataset": {
                    "bad_dataset": ["file1", {"dataset": ["file2"]}]}}},
    }, False),  # Same issue but further down the tree
    ({"valid_container": {"dataset": ["file1"], "dataset2": ["file2"]}}, True),  # One valid container
    ({
        "valid_container": {"dataset": ["file1"], "dataset2": ["file2"]},
        "valid_container2": {"dataset": ["file1"], "dataset2": ["file2"]}
    }, True),  # Two valid containers
    ({
        "valid_container": {
            "nested_valid_container": {"dataset": ["file1"]},
            "nested_valid_container2": {"dataset": ["file2"]}}
    },  # Nested valid containers
        True
    )
],
    ids=[
        "Single invalid container",
        "Invalid container container paired with valid container",
        "Nested invalid container",
        "Valid single container",
        "Multiple valid container",
        "Nested valid containers"
])
def test_upload_recursive(structure, is_valid_container, rse_factory, scope, upload_client):
    """
    Ensure logic for recursive uploads.

    Containers only contain collection DIDs of one type, and never files. Datasets only contain files or other datasets, but not at the same time.
    """

    def recursive_create(dir_path, contents):
        for value in contents.values():
            directory = did_name_generator()

            if isinstance(value, list):
                for item in value:
                    if isinstance(item, str):
                        file_path = os.path.join(dir_path, directory,  ''.join(choice(ascii_uppercase) for _ in range(5)))
                        os.makedirs(os.path.dirname(file_path), exist_ok=True)
                        execute(f'dd if=/dev/urandom of={file_path} count={2} bs=1')

                    elif isinstance(item, dict):
                        subdir = did_name_generator()
                        sub_dir_path = os.path.join(dir_path, directory, subdir)
                        os.makedirs(sub_dir_path, exist_ok=True)
                        recursive_create(sub_dir_path, item)
                    else:
                        raise ValueError(f"Invalid item type: {type(item)} in structure")

            else:
                # item is a subdirectory
                sub_dir_path = os.path.join(dir_path, directory)
                os.makedirs(sub_dir_path, exist_ok=True)
                recursive_create(sub_dir_path, value)

    with TemporaryDirectory() as tmp_dir:

        recursive_create(tmp_dir, structure)

        rse, _ = rse_factory.make_posix_rse()
        items = {
            'path': tmp_dir,
            'rse': rse,
            'did_scope': scope,
            'did_name': os.path.basename(tmp_dir),
            'recursive': True,
        }
        if not is_valid_container:
            with pytest.raises(InputValidationError):
                upload_client._collect_files_recursive(items)
        else:
            upload_client._collect_files_recursive(items)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy

import pytest

from rucio.client.didclient import DIDClient
from rucio.common.exception import KeyNotFound
from rucio.common.utils import generate_uuid
from rucio.core.did import add_did, delete_dids, get_metadata_bulk, set_dids_metadata_bulk, set_metadata_bulk
from rucio.core.did_meta_plugins import get_metadata, list_dids, set_metadata
from rucio.core.did_meta_plugins.elasticsearch_meta import ElasticDidMeta
from rucio.core.did_meta_plugins.mongo_meta import MongoDidMeta
from rucio.core.did_meta_plugins.postgres_meta import ExternalPostgresJSONDidMeta
from rucio.db.sqla.util import json_implemented
from rucio.tests.common import did_name_generator, skip_rse_tests_with_accounts


def skip_without_json():
    if not json_implemented():
        pytest.skip("JSON support is not implemented in this database")


class TestDidMetaDidColumn:

    @pytest.mark.dirty
    def test_add_did_meta(self, mock_scope, root_account):
        """ DID Meta (Hardcoded): Add DID meta """
        did_name = did_name_generator('dataset')
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        set_metadata(scope=mock_scope, name=did_name, key='project', value='data12_8TeV')
        assert get_metadata(scope=mock_scope, name=did_name)['project'] == 'data12_8TeV'

    @pytest.mark.dirty
    def test_get_did_meta(self, mock_scope, root_account):
        """ DID Meta (Hardcoded): Get DID meta """
        did_name = did_name_generator('dataset')
        dataset_meta = {'project': 'data12_8TeV'}
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', meta=dataset_meta, account=root_account)
        assert get_metadata(scope=mock_scope, name=did_name)['project'] == 'data12_8TeV'

    @pytest.mark.dirty
    def test_list_did_meta(self, mock_scope, root_account):
        """ DID Meta (Hardcoded): List DID meta """
        dsns = []
        tmp_dsn1 = did_name_generator('dataset')

        dsns.append(tmp_dsn1)

        dataset_meta = {'project': 'data12_8TeV',
                        'run_number': 400000,
                        'stream_name': 'physics_CosmicCalo',
                        'prod_step': 'merge',
                        'datatype': 'NTUP_TRIG',
                        'version': 'f392_m920',
                        }

        add_did(scope=mock_scope, name=tmp_dsn1, did_type="DATASET", account=root_account, meta=dataset_meta)

        tmp_dsn2 = did_name_generator('dataset')
        dsns.append(tmp_dsn2)
        dataset_meta['run_number'] = 400001
        add_did(scope=mock_scope, name=tmp_dsn2, did_type="DATASET", account=root_account, meta=dataset_meta)

        tmp_dsn3 = did_name_generator('dataset')
        dsns.append(tmp_dsn3)
        dataset_meta['stream_name'] = 'physics_Egamma'
        dataset_meta['datatype'] = 'NTUP_SMWZ'
        add_did(scope=mock_scope, name=tmp_dsn3, did_type="DATASET", account=root_account, meta=dataset_meta)

        dids = list_dids(mock_scope, {'project': 'data12_8TeV', 'version': 'f392_m920'})
        results = []
        for d in dids:
            results.append(d)
        for dsn in dsns:
            assert dsn in results
        dsns.remove(tmp_dsn1)

        dids = list_dids(mock_scope, {'project': 'data12_8TeV', 'run_number': 400001})
        results = []
        for d in dids:
            results.append(d)
        for dsn in dsns:
            assert dsn in results
        dsns.remove(tmp_dsn2)

        dids = list_dids(mock_scope, {'project': 'data12_8TeV', 'stream_name': 'physics_Egamma', 'datatype': 'NTUP_SMWZ'})
        results = []
        for d in dids:
            results.append(d)
        for dsn in dsns:
            assert dsn in results

        # with pytest.raises(KeyNotFound):
        #     list_dids(tmp_scope, {'NotReallyAKey': 'NotReallyAValue'})


class TestDidMetaJSON:

    @pytest.mark.dirty
    def test_add_did_meta(self, mock_scope, root_account):
        """ DID Meta (JSON): Add DID meta """
        skip_without_json()

        did_name = did_name_generator('dataset')
        meta_key = 'my_key_%s' % generate_uuid()
        meta_value = 'my_value_%s' % generate_uuid()
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        set_metadata(scope=mock_scope, name=did_name, key=meta_key, value=meta_value)

        meta = get_metadata(scope=mock_scope, name=did_name, plugin='JSON')
        assert meta[meta_key] == meta_value

    @pytest.mark.dirty
    def test_get_metadata_bulk(self, mock_scope, root_account):
        skip_without_json()

        dids = []
        expected_meta = []

        meta_key = 'data_product_id'
        for _ in range(5):
            did_name = did_name_generator('dataset')
            meta_value = generate_uuid()
            add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
            set_metadata(scope=mock_scope, name=did_name, key=meta_key, value=meta_value)
            dids.append({'scope': mock_scope, 'name': did_name})
            expected_meta.append({'scope': mock_scope, 'name': did_name, meta_key: meta_value})

        meta = list(get_metadata_bulk(dids, plugin="JSON", inherit=True))
        assert meta == expected_meta

    @pytest.mark.dirty
    def test_get_metadata(self, mock_scope, root_account):
        """ DID Meta (JSON): Get DID meta """
        skip_without_json()

        did_name = did_name_generator('dataset')
        meta_key = 'my_key_%s' % generate_uuid()
        meta_value = 'my_value_%s' % generate_uuid()
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        set_metadata(scope=mock_scope, name=did_name, key=meta_key, value=meta_value)
        assert get_metadata(scope=mock_scope, name=did_name, plugin='JSON')[meta_key] == meta_value

    @pytest.mark.dirty
    def test_list_did_meta(self, mock_scope, root_account):
        """ DID Meta (JSON): List DID meta """
        skip_without_json()

        meta_key1 = 'my_key_%s' % generate_uuid()
        meta_key2 = 'my_key_%s' % generate_uuid()
        meta_value1 = 'my_value_%s' % generate_uuid()
        meta_value2 = 'my_value_%s' % generate_uuid()

        tmp_dsn1 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn1, did_type="DATASET", account=root_account)
        set_metadata(scope=mock_scope, name=tmp_dsn1, key=meta_key1, value=meta_value1)

        tmp_dsn2 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn2, did_type="DATASET", account=root_account)
        set_metadata(scope=mock_scope, name=tmp_dsn2, key=meta_key1, value=meta_value2)

        tmp_dsn3 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn3, did_type="DATASET", account=root_account)
        set_metadata(scope=mock_scope, name=tmp_dsn3, key=meta_key2, value=meta_value1)

        tmp_dsn4 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn4, did_type="DATASET", account=root_account)
        set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key1, value=meta_value1)
        set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key2, value=meta_value2)

        dids = list_dids(mock_scope, {meta_key1: meta_value1})
        results = sorted(list(dids))

        assert len(results) == 2
        # assert sorted([{'scope': tmp_scope, 'name': tmp_dsn1}, {'scope': tmp_scope, 'name': tmp_dsn4}]) == sorted(results)
        expected = sorted([tmp_dsn1, tmp_dsn4])
        assert expected == results

        dids = list_dids(mock_scope, {meta_key1: meta_value2})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': str(tmp_dsn2)}] == results
        assert [tmp_dsn2] == results

        dids = list_dids(mock_scope, {meta_key2: meta_value1})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': tmp_dsn3}] == results
        assert [tmp_dsn3] == results

        dids = list_dids(mock_scope, {meta_key1: meta_value1, meta_key2: meta_value2})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': tmp_dsn4}] == results
        assert [tmp_dsn4] == results


@pytest.fixture(params=["with-auth", "no-auth"])
def mongo_meta(request):
    if request.param == "with-auth":
        return MongoDidMeta(
            host='mongo',
            port=27017,
            db='test_db',
            collection='test_collection',
            user="rucio",
            password="mongo-meta",
        )
    else:
        return MongoDidMeta(
            host='mongo-noauth',
            port=27017,
            db='test_db',
            collection='test_collection',
        )


@skip_rse_tests_with_accounts
class TestDidMetaMongo:

    @pytest.mark.dirty
    def test_set_get_metadata(self, mock_scope, root_account, mongo_meta):
        """ DID Meta (MONGO): Get/set DID meta """

        did_name = did_name_generator('dataset')
        meta_key = 'my_key_%s' % generate_uuid()
        meta_value = 'my_value_%s' % generate_uuid()
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        mongo_meta.set_metadata(scope=mock_scope, name=did_name, key=meta_key, value=meta_value)
        assert mongo_meta.get_metadata(scope=mock_scope, name=did_name)[meta_key] == meta_value

    @pytest.mark.dirty
    def test_list_did_meta(self, mock_scope, root_account, mongo_meta):
        """ DID Meta (MONGO): List DID meta """

        meta_key1 = 'my_key_%s' % generate_uuid()
        meta_key2 = 'my_key_%s' % generate_uuid()
        meta_value1 = 'my_value_%s' % generate_uuid()
        meta_value2 = 'my_value_%s' % generate_uuid()

        tmp_dsn1 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn1, did_type="DATASET", account=root_account)
        mongo_meta.set_metadata(scope=mock_scope, name=tmp_dsn1, key=meta_key1, value=meta_value1)

        tmp_dsn2 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn2, did_type="DATASET", account=root_account)
        mongo_meta.set_metadata(scope=mock_scope, name=tmp_dsn2, key=meta_key1, value=meta_value2)

        tmp_dsn3 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn3, did_type="DATASET", account=root_account)
        mongo_meta.set_metadata(scope=mock_scope, name=tmp_dsn3, key=meta_key2, value=meta_value1)

        tmp_dsn4 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn4, did_type="DATASET", account=root_account)
        mongo_meta.set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key1, value=meta_value1)
        mongo_meta.set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key2, value=meta_value2)

        dids = mongo_meta.list_dids(mock_scope, {meta_key1: meta_value1})
        results = sorted(list(dids))

        assert len(results) == 2
        # assert sorted([{'scope': tmp_scope, 'name': tmp_dsn1}, {'scope': tmp_scope, 'name': tmp_dsn4}]) == sorted(results)
        expected = sorted([tmp_dsn1, tmp_dsn4])
        assert expected == results

        dids = mongo_meta.list_dids(mock_scope, {meta_key1: meta_value2})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': str(tmp_dsn2)}] == results
        assert [tmp_dsn2] == results

        dids = mongo_meta.list_dids(mock_scope, {meta_key2: meta_value1})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': tmp_dsn3}] == results
        assert [tmp_dsn3] == results

        dids = mongo_meta.list_dids(mock_scope, {meta_key1: meta_value1, meta_key2: meta_value2})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': tmp_dsn4}] == results
        assert [tmp_dsn4] == results


@pytest.fixture
def elastic_meta():
    return ElasticDidMeta(
            hosts=['http://elasticsearch_meta:9200'],
            user="elastic",
            password="rucio",
        )


@skip_rse_tests_with_accounts
class TestDidMetaElastic:

    @pytest.mark.dirty
    def test_set_get_metadata(self, mock_scope, root_account, elastic_meta):
        """ DID Meta (ELASTIC): Get/set DID meta """
        did_name = did_name_generator('dataset')
        meta_key = 'my_key_%s' % generate_uuid()
        meta_value = 'my_value_%s' % generate_uuid()
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        elastic_meta.set_metadata(scope=mock_scope, name=did_name, key=meta_key, value=meta_value)
        assert elastic_meta.get_metadata(scope=mock_scope, name=did_name)[meta_key] == meta_value

    @pytest.mark.dirty
    def test_delete_metadata(self, mock_scope, root_account, elastic_meta):
        """ DID Meta (ELASTIC): Deletes metadata key """

        meta_key1 = 'my_key_%s' % generate_uuid()
        meta_key2 = 'my_key_%s' % generate_uuid()
        meta_value1 = 'my_value_%s' % generate_uuid()
        meta_value2 = 'my_value_%s' % generate_uuid()

        tmp_dsn1 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn1, did_type="DATASET", account=root_account)
        meta = {meta_key1: meta_value1, meta_key2: meta_value2}
        elastic_meta.set_metadata_bulk(scope=mock_scope, name=tmp_dsn1, meta=meta)
        metadata = elastic_meta.get_metadata(scope=mock_scope, name=tmp_dsn1)
        assert metadata[meta_key1] == meta_value1
        assert metadata[meta_key2] == meta_value2

        elastic_meta.delete_metadata(scope=mock_scope, name=tmp_dsn1, key=meta_key2)

        metadata = elastic_meta.get_metadata(scope=mock_scope, name=tmp_dsn1)
        assert metadata[meta_key1] == meta_value1
        assert meta_key2 not in metadata

    @pytest.mark.dirty
    def test_list_did_meta(self, mock_scope, root_account, elastic_meta):
        """ DID Meta (ELASTIC): List DID meta """

        meta_key1 = 'my_key_%s' % generate_uuid()
        meta_key2 = 'my_key_%s' % generate_uuid()
        meta_value1 = 'my_value_%s' % generate_uuid()
        meta_value2 = 'my_value_%s' % generate_uuid()

        tmp_dsn1 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn1, did_type="DATASET", account=root_account)
        elastic_meta.set_metadata(scope=mock_scope, name=tmp_dsn1, key=meta_key1, value=meta_value1)

        tmp_dsn2 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn2, did_type="DATASET", account=root_account)
        elastic_meta.set_metadata(scope=mock_scope, name=tmp_dsn2, key=meta_key1, value=meta_value2)

        tmp_dsn3 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn3, did_type="DATASET", account=root_account)
        elastic_meta.set_metadata(scope=mock_scope, name=tmp_dsn3, key=meta_key2, value=meta_value1)
        elastic_meta.get_metadata(scope=mock_scope, name=tmp_dsn3)

        tmp_dsn4 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn4, did_type="DATASET", account=root_account)
        elastic_meta.set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key1, value=meta_value1)
        elastic_meta.set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key2, value=meta_value2)

        dids = elastic_meta.list_dids(mock_scope, {meta_key1: meta_value1})
        results = sorted(list(dids))
        assert len(results) == 2
        expected = sorted([tmp_dsn1, tmp_dsn4])
        assert expected == results

        results = list(elastic_meta.list_dids(mock_scope, {meta_key1: meta_value2}))
        assert len(results) == 1
        assert [tmp_dsn2] == results

        results = list(elastic_meta.list_dids(mock_scope, {meta_key2: meta_value1}))
        assert len(results) == 1
        assert [tmp_dsn3] == results

        results = list(elastic_meta.list_dids(mock_scope, {meta_key1: meta_value1, meta_key2: meta_value2}))
        assert len(results) == 1
        assert [tmp_dsn4] == results


@pytest.fixture
def postgres_json_meta():
    return ExternalPostgresJSONDidMeta(
        host='postgres',
        port=5433,
        db='metadata',
        user='rucio',
        password='secret',
        db_schema='public',
        table='dids',
        table_is_managed=True,
        table_column_vo='vo',
        table_column_scope='scope',
        table_column_name='name',
        table_column_data='data',
    )


@pytest.mark.noparallel(reason='race condition on try-create table')
@skip_rse_tests_with_accounts
class TestDidMetaExternalPostgresJSON:

    @pytest.mark.dirty
    def test_set_get_metadata(self, mock_scope, root_account, postgres_json_meta):
        """ DID Meta (POSTGRES_JSON): Get/set DID meta """

        did_name = did_name_generator('dataset')
        meta_key = 'my_key_%s' % generate_uuid()
        meta_value = 'my_value_%s' % generate_uuid()
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        postgres_json_meta.set_metadata(scope=mock_scope, name=did_name, key=meta_key, value=meta_value)
        assert postgres_json_meta.get_metadata(scope=mock_scope, name=did_name)[meta_key] == meta_value

    @pytest.mark.dirty
    def test_delete_metadata(self, mock_scope, root_account, postgres_json_meta):
        """ DID Meta (POSTGRES_JSON): delete DID meta """

        did_name = did_name_generator('dataset')
        meta_key1 = 'my_key_%s' % generate_uuid()
        meta_value1 = 'my_value_%s' % generate_uuid()
        meta_key2 = 'my_key_%s' % generate_uuid()
        meta_value2 = 'my_value_%s' % generate_uuid()
        add_did(scope=mock_scope, name=did_name, did_type='DATASET', account=root_account)
        postgres_json_meta.set_metadata(scope=mock_scope, name=did_name, key=meta_key1, value=meta_value1)
        postgres_json_meta.set_metadata(scope=mock_scope, name=did_name, key=meta_key2, value=meta_value2)
        
        assert postgres_json_meta.get_metadata(scope=mock_scope, name=did_name)[meta_key1] == meta_value1
        assert postgres_json_meta.get_metadata(scope=mock_scope, name=did_name)[meta_key2] == meta_value2
       
        postgres_json_meta.delete_metadata(scope=mock_scope, name=did_name, key=meta_key2)
        metadata = postgres_json_meta.get_metadata(scope=mock_scope, name=did_name)
        assert metadata[meta_key1] == meta_value1
        assert meta_key2 not in metadata

    @pytest.mark.dirty
    def test_list_did_meta(self, mock_scope, root_account, postgres_json_meta):
        """ DID Meta (POSTGRES_JSON): List DID meta """

        meta_key1 = 'my_key_%s' % generate_uuid()
        meta_key2 = 'my_key_%s' % generate_uuid()
        meta_value1 = 'my_value_%s' % generate_uuid()
        meta_value2 = 'my_value_%s' % generate_uuid()

        tmp_dsn1 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn1, did_type="DATASET", account=root_account)
        postgres_json_meta.set_metadata(scope=mock_scope, name=tmp_dsn1, key=meta_key1, value=meta_value1)

        tmp_dsn2 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn2, did_type="DATASET", account=root_account)
        postgres_json_meta.set_metadata(scope=mock_scope, name=tmp_dsn2, key=meta_key1, value=meta_value2)

        tmp_dsn3 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn3, did_type="DATASET", account=root_account)
        postgres_json_meta.set_metadata(scope=mock_scope, name=tmp_dsn3, key=meta_key2, value=meta_value1)

        tmp_dsn4 = did_name_generator('dataset')
        add_did(scope=mock_scope, name=tmp_dsn4, did_type="DATASET", account=root_account)
        postgres_json_meta.set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key1, value=meta_value1)
        postgres_json_meta.set_metadata(scope=mock_scope, name=tmp_dsn4, key=meta_key2, value=meta_value2)

        dids = postgres_json_meta.list_dids(mock_scope, {meta_key1: meta_value1})
        results = sorted(list(dids))

        assert len(results) == 2
        # assert sorted([{'scope': tmp_scope, 'name': tmp_dsn1}, {'scope': tmp_scope, 'name': tmp_dsn4}]) == sorted(results)
        expected = sorted([tmp_dsn1, tmp_dsn4])
        assert expected == results

        dids = postgres_json_meta.list_dids(mock_scope, {meta_key1: meta_value2})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': str(tmp_dsn2)}] == results
        assert [tmp_dsn2] == results

        dids = postgres_json_meta.list_dids(mock_scope, {meta_key2: meta_value1})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': tmp_dsn3}] == results
        assert [tmp_dsn3] == results

        dids = postgres_json_meta.list_dids(mock_scope, {meta_key1: meta_value1, meta_key2: meta_value2})
        results = []
        for d in dids:
            results.append(d)
        assert len(results) == 1
        # assert [{'scope': (tmp_scope), 'name': tmp_dsn4}] == results
        assert [tmp_dsn4] == results


class TestDidMetaClient:

    @pytest.mark.dirty
    def test_set_metadata(self, mock_scope, did_client, db_session):
        """ META (CLIENTS): Adds a fully set json column to a DID, updates if some keys present """
        tmp_name = did_name_generator('dataset')
        scope = mock_scope.external
        did_client.add_did(scope=scope, name=tmp_name, did_type="DATASET")

        # Test JSON case
        if json_implemented(session=db_session):
            # data1 = ["key1": "value_" + str(generate_uuid()), "key2": "value_" + str(generate_uuid()), "key3": "value_" + str(generate_uuid())]
            value1 = "value_" + str(generate_uuid())
            value2 = "value_" + str(generate_uuid())
            value3 = "value_" + str(generate_uuid())
            did_client.set_metadata(scope=scope, name=tmp_name, key="key1", value=value1)
            did_client.set_metadata(scope=scope, name=tmp_name, key="key2", value=value2)
            did_client.set_metadata(scope=scope, name=tmp_name, key="key3", value=value3)

            metadata = did_client.get_metadata(scope=scope, name=tmp_name, plugin="JSON")

            assert len(metadata) == 3
            assert metadata['key1'] == value1
            assert metadata['key2'] == value2
            assert metadata['key3'] == value3

        # Test DID_COLUMNS case
        did_client.set_metadata(scope=scope, name=tmp_name, key='project', value='data12_12TeV')
        assert did_client.get_metadata(scope=scope, name=tmp_name)['project'] == 'data12_12TeV'

    @pytest.mark.dirty
    def test_delete_metadata(self, mock_scope, did_client):
        """ META (CLIENTS): Deletes metadata key """
        skip_without_json()
        scope = mock_scope.external
        tmp_name = did_name_generator('dataset')
        did_client.add_did(scope=scope, name=tmp_name, did_type="DATASET")

        value1 = "value_" + str(generate_uuid())
        value2 = "value_" + str(generate_uuid())
        value3 = "value_" + str(generate_uuid())

        did_client.set_metadata(scope=scope, name=tmp_name, key="key1", value=value1)
        did_client.set_metadata(scope=scope, name=tmp_name, key="key2", value=value2)
        did_client.set_metadata(scope=scope, name=tmp_name, key="key3", value=value3)

        did_client.delete_metadata(scope=scope, name=tmp_name, key='key2')

        metadata = did_client.get_metadata(scope=scope, name=tmp_name, plugin="JSON")
        assert len(metadata) == 2
        assert metadata['key1'] == value1
        assert metadata['key3'] == value3
        with pytest.raises(KeyNotFound):
            did_client.delete_metadata(scope=scope, name=tmp_name, key="key9")

    @pytest.mark.dirty
    def test_get_metadata(self, mock_scope, did_client, db_session):
        """ META (CLIENTS): Gets all metadata for the given DID """
        tmp_name = did_name_generator('dataset')
        scope = mock_scope.external
        did_client.add_did(scope=scope, name=tmp_name, did_type="DATASET")

        # Test JSON case
        if json_implemented(session=db_session):
            value1 = "value_" + str(generate_uuid())
            value2 = "value_" + str(generate_uuid())

            did_client.set_metadata(scope=scope, name=tmp_name, key="key1", value=value1)
            did_client.set_metadata(scope=scope, name=tmp_name, key="key2", value=value2)

            metadata = did_client.get_metadata(scope=scope, name=tmp_name, plugin="JSON")

            assert len(metadata) == 2
            assert metadata['key1'] == value1
            assert metadata['key2'] == value2

        # Test DID_COLUMNS case
        did_client.set_metadata(scope=scope, name=tmp_name, key='project', value='data12_14TeV')
        assert did_client.get_metadata(scope=scope, name=tmp_name)['project'] == 'data12_14TeV'

        # Test Mixed case
        if json_implemented(session=db_session):
            all_metadata = did_client.get_metadata(scope=scope, name=tmp_name, plugin="ALL")
            assert all_metadata['key1'] == value1
            assert all_metadata['key2'] == value2
            assert all_metadata['project'] == "data12_14TeV"

    @pytest.mark.dirty
    @pytest.mark.noparallel(reason='fails when run in parallel')
    def test_list_dids(self, did_client, db_session):
        """ META (CLIENTS): Get all DIDs matching the values of the provided metadata keys """

        # Test DID Columns use case
        dsns = []
        tmp_scope = 'mock'
        tmp_dsn1 = did_name_generator('dataset')
        dsns.append(tmp_dsn1)

        dataset_meta = {'project': 'data12_8TeV',
                        'run_number': 400000,
                        'stream_name': 'physics_CosmicCalo',
                        'prod_step': 'merge',
                        'datatype': 'NTUP_TRIG',
                        'version': 'f392_m920',
                        }
        did_client.add_dataset(scope=tmp_scope, name=tmp_dsn1, meta=dataset_meta)
        tmp_dsn2 = did_name_generator('dataset')
        dsns.append(tmp_dsn2)
        dataset_meta['run_number'] = 400001
        did_client.add_dataset(scope=tmp_scope, name=tmp_dsn2, meta=dataset_meta)

        tmp_dsn3 = did_name_generator('dataset')
        dsns.append(tmp_dsn3)
        dataset_meta['stream_name'] = 'physics_Egamma'
        dataset_meta['datatype'] = 'NTUP_SMWZ'
        did_client.add_dataset(scope=tmp_scope, name=tmp_dsn3, meta=dataset_meta)

        dids = did_client.list_dids(tmp_scope, {'project': 'data12_8TeV', 'version': 'f392_m920'})
        results = []
        for d in dids:
            results.append(d)
        for dsn in dsns:
            assert dsn in results
        dsns.remove(tmp_dsn1)

        dids = did_client.list_dids(tmp_scope, {'project': 'data12_8TeV', 'run_number': 400001})
        results = []
        for d in dids:
            results.append(d)
        for dsn in dsns:
            assert dsn in results
        dsns.remove(tmp_dsn2)

        dids = did_client.list_dids(tmp_scope, {'project': 'data12_8TeV', 'stream_name': 'physics_Egamma', 'datatype': 'NTUP_SMWZ'})
        results = []
        for d in dids:
            results.append(d)
        for dsn in dsns:
            assert dsn in results

        # Test JSON use case
        if json_implemented(session=db_session):
            did1 = did_name_generator('dataset')
            did2 = did_name_generator('dataset')
            did3 = did_name_generator('dataset')
            did4 = did_name_generator('dataset')

            key1 = 'key_1_%s' % generate_uuid()
            key2 = 'key_2_%s' % generate_uuid()
            key3 = 'key_3_%s' % generate_uuid()

            value1 = 'value_1_%s' % generate_uuid()
            value2 = 'value_2_%s' % generate_uuid()
            value3 = 'value_3_%s' % generate_uuid()
            value_not_1 = 'value_not_1_%s' % generate_uuid()
            value_not_2 = 'value_not_1_%s' % generate_uuid()
            value_unique = 'value_unique_%s' % generate_uuid()

            did_client.add_did(scope=tmp_scope, name=did1, did_type="DATASET")
            did_client.add_did(scope=tmp_scope, name=did2, did_type="DATASET")
            did_client.add_did(scope=tmp_scope, name=did3, did_type="DATASET")
            did_client.add_did(scope=tmp_scope, name=did4, did_type="DATASET")

            did_client.set_metadata(scope=tmp_scope, name=did1, key=key1, value=value1)
            did_client.set_metadata(scope=tmp_scope, name=did1, key=key2, value=value2)

            did_client.set_metadata(scope=tmp_scope, name=did2, key=key1, value=value1)
            did_client.set_metadata(scope=tmp_scope, name=did2, key=key2, value=value_not_2)
            did_client.set_metadata(scope=tmp_scope, name=did2, key=key3, value=value3)

            did_client.set_metadata(scope=tmp_scope, name=did3, key=key1, value=value_not_1)
            did_client.set_metadata(scope=tmp_scope, name=did3, key=key2, value=value2)
            did_client.set_metadata(scope=tmp_scope, name=did3, key=key3, value=value3)

            did_client.set_metadata(scope=tmp_scope, name=did4, key=key1, value=value1)
            did_client.set_metadata(scope=tmp_scope, name=did4, key=key2, value=value2)
            did_client.set_metadata(scope=tmp_scope, name=did4, key=key3, value=value_unique)

            # Key not there
            dids = did_client.list_dids(tmp_scope, {'key45': 'value'})
            results = []
            for d in dids:
                results.append(d)
            assert len(results) == 0

            # Value not there
            dids = did_client.list_dids(tmp_scope, {key1: 'value_not_there'})
            results = []
            for d in dids:
                results.append(d)
            assert len(results) == 0

            # key1 = value1
            dids = did_client.list_dids(tmp_scope, {key1: value1})
            results = []
            for d in dids:
                results.append(d)
            assert len(results) == 3
            assert did1 in results
            assert did2 in results
            assert did4 in results

            # key1, key2
            dids = did_client.list_dids(tmp_scope, {key1: value1, key2: value2})
            results = []
            for d in dids:
                results.append(d)
            assert len(results) == 2
            assert did1 in results
            assert did4 in results

            # key1, key2, key 3
            dids = did_client.list_dids(tmp_scope, {key1: value1, key2: value2, key3: value3})
            results = []
            for d in dids:
                results.append(d)
            assert len(results) == 0

            # key3 = unique value
            dids = did_client.list_dids(tmp_scope, {key3: value_unique})
            results = []
            for d in dids:
                results.append(d)
            assert len(results) == 1
            assert did4 in results


@pytest.fixture
def testdid(vo, mock_scope, root_account):
    did_name = did_name_generator('dataset')
    didtype = 'DATASET'

    add_did(scope=mock_scope, name=did_name, did_type=didtype, account=root_account)
    yield {'name': did_name, 'scope': mock_scope}
    delete_dids(dids=[{'name': did_name, 'scope': mock_scope, 'did_type': didtype, 'purge_replicas': True}], account=root_account)


def test_did_set_metadata_bulk_single(testdid):
    """ DID (CORE): Test setting metadata in bulk with a single key-value pair """
    skip_without_json()

    testkey = 'testkey'
    testmeta = {testkey: 'testvalue'}

    set_metadata_bulk(meta=testmeta, recursive=False, **testdid)
    meta = get_metadata(plugin="ALL", **testdid)
    print('Metadata:', meta)

    assert testkey in meta and meta[testkey] == testmeta[testkey]


def test_did_set_metadata_bulk_multi(testdid):
    """ DID (CORE): Test setting metadata in bulk with multiple key-values """
    skip_without_json()

    testkeys = list(map(lambda i: 'testkey' + str(i), range(3)))
    testmeta = {key: key + 'value' for key in testkeys}
    # let two keys have the same value
    testmeta[testkeys[1]] = testmeta[testkeys[0]]

    set_metadata_bulk(meta=testmeta, recursive=False, **testdid)
    meta = get_metadata(plugin="ALL", **testdid)
    print('Metadata:', meta)

    for testkey in testkeys:
        assert testkey in meta and meta[testkey] == testmeta[testkey]


def test_set_dids_metadata_bulk_multi(did_factory):
    """ DID (CORE): Test setting metadata in bulk with multiple key-values on multiple DIDs"""
    skip_without_json()
    nb_dids = 5
    dids = [did_factory.make_dataset() for _ in range(nb_dids)]

    for did in dids:
        testkeys = list(map(lambda i: 'testkey' + generate_uuid(), range(3)))
        testmeta = {key: key + 'value' for key in testkeys}
        did['meta'] = testmeta
    print(dids)

    set_dids_metadata_bulk(dids=dids, recursive=False)
    for did in dids:
        testmeta = did['meta']
        print('Metadata:', testmeta)
        meta = get_metadata(plugin="ALL", scope=did['scope'], name=did['name'])
        print('Metadata:', meta)
        for testkey in testmeta:
            assert testkey in meta and meta[testkey] == testmeta[testkey]


def test_did_set_metadata_bulk_multi_client(testdid):
    """ DID (CLIENT): Test setting metadata in bulk with multiple key-values """
    skip_without_json()

    testkeys = list(map(lambda i: 'testkey' + str(i), range(3)))
    testmeta = {key: key + 'value' for key in testkeys}
    # let two keys have the same value
    testmeta[testkeys[1]] = testmeta[testkeys[0]]

    didclient = DIDClient()
    external_testdid = testdid.copy()
    external_testdid['scope'] = testdid['scope'].external
    result = didclient.set_metadata_bulk(meta=testmeta, recursive=False, **external_testdid)
    assert result is True

    meta = get_metadata(plugin="ALL", **testdid)
    print('Metadata:', meta)

    for testkey in testkeys:
        assert testkey in meta and meta[testkey] == testmeta[testkey]


def test_set_dids_metadata_bulk_multi_client(did_factory, rucio_client):
    """ DID (CLIENT): Test setting metadata in bulk with multiple key-values on multiple DIDs"""
    skip_without_json()
    nb_dids = 5
    dids = [did_factory.make_dataset() for _ in range(nb_dids)]
    for did in dids:
        testkeys = list(map(lambda i: 'testkey' + generate_uuid(), range(3)))
        testmeta = {key: key + 'value' for key in testkeys}
        did['meta'] = testmeta

    external_testdids = deepcopy(dids)
    for did in external_testdids:
        did['scope'] = did['scope'].external
    print(dids)
    print(external_testdids)

    result = rucio_client.set_dids_metadata_bulk(dids=external_testdids, recursive=False)
    assert result is True

    for did in dids:
        testmeta = did['meta']
        print('Metadata:', testmeta)
        meta = get_metadata(plugin="ALL", scope=did['scope'], name=did['name'])
        print('Metadata:', meta)
        for testkey in testmeta:
            assert testkey in meta and meta[testkey] == testmeta[testkey]



#!/usr/bin/env python
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Blueprint, make_response, render_template, request

from rucio.common.config import config_get_bool
from rucio.common.constants import DEFAULT_VO, HTTPMethod
from rucio.common.policy import get_policy
from rucio.gateway.authentication import get_auth_token_x509
from rucio.web.rest.flaskapi.v1.common import generate_http_error_flask
from rucio.web.ui.flask.common.utils import AUTH_ISSUERS, SAML_SUPPORT, USERPASS_SUPPORT, authenticate, finalize_auth, get_token, oidc_auth, saml_auth, userpass_auth, x509token_auth

MULTI_VO = config_get_bool('common', 'multi_vo', raise_exception=False, default=False)
POLICY = get_policy()
ATLAS_URLS = ()
OTHER_URLS = ()


def auth():
    auth_type = request.cookies.get('x-rucio-auth-type')
    if str(auth_type).lower() == 'x509':
        token = get_token(get_auth_token_x509)
        if token:
            response = make_response()
            response.headers['X-Rucio-Auth-Token'] = token
            print("auth()")
            print(token)
            return response
        else:
            return generate_http_error_flask(401, 'CannotAuthenticate', 'Cannot get token')
    else:
        return render_template('select_login_method.html', oidc_issuers=AUTH_ISSUERS, saml_support=SAML_SUPPORT,
                               userpass_support=USERPASS_SUPPORT)


def login():
    if request.method == HTTPMethod.GET.value:
        account = request.args.get('account')
        vo = request.args.get('vo')
        return render_template('login.html', account=account, vo=vo, userpass_support=USERPASS_SUPPORT)
    if request.method == HTTPMethod.POST.value:
        return userpass_auth()


def oidc():
    account = request.args.get('account')
    issuer = request.args.get('issuer')
    vo = request.args.get('vo')
    if not MULTI_VO:
        vo = DEFAULT_VO

    if not issuer:
        return generate_http_error_flask(401, 'CannotAuthenticate', 'Cannot get token OIDC auth url from the server.')

    return oidc_auth(account, issuer, vo)


def oidc_final():
    session_token = request.cookies.get('x-rucio-auth-token')
    return finalize_auth(session_token, 'OIDC')


def saml():
    return saml_auth(request.method)


def x509():
    return x509token_auth()


AUTH_URLS = (
    ('/auth', 'auth', auth, [HTTPMethod.GET]),
    ('/login', 'login', login, [HTTPMethod.GET, HTTPMethod.POST]),
    ('/oidc', 'oidc', oidc, [HTTPMethod.GET]),
    ('/oidc_final', 'oidc_final', oidc_final, [HTTPMethod.GET]),
    ('/saml', 'saml', saml, [HTTPMethod.GET, HTTPMethod.POST]),
    ('/x509', 'x509', x509, [HTTPMethod.GET])
)

COMMON_URLS = (
    ('/account_rse_usage', 'account_rse_usage', 'Account RSE Usage'),
    ('/account', 'account', 'Account'),
    ('/bad_replicas', 'bad_replicas', 'Bad Replicas'),
    ('/bad_replicas/summary', 'bad_replicas_summary', 'Bad Replica Summary'),
    ('/did', 'did', 'Data Identifier'),
    ('/heartbeats', 'heartbeats', 'Heartbeats'),
    ('/lifetime_exception', 'lifetime_exception', 'Lifetime Exception'),
    ('/list_lifetime_exceptions', 'list_lifetime_exceptions', 'Lifetime Exception'),
    ('/list_accounts', 'accounts', 'Accounts'),
    ('/r2d2/approve', 'approve_rules', 'Rules in Approval State'),
    ('/r2d2/request', 'request_rule', 'Rucio Rule Definition Droid - Request Rule'),
    ('/r2d2/manage_quota', 'rse_account_usage', 'Manage Quota'),
    ('/r2d2', 'list_rules', 'Rucio Rule Definition Droid - List Rules'),
    ('/rse', 'rse', 'RSE Info'),
    ('/rse/protocol/add', 'rse_add_protocol', 'RSE Protocol'),
    ('/rses', 'rses', 'RSEs'),
    ('/rses/add', 'add_rse', 'Add RSE'),
    ('/rse_usage', 'rse_usage', 'RSE Usage'),
    ('/rse_locks', 'rse_locks', 'RSE Locks'),
    ('/rule', 'rule', 'Rule'),
    ('/rules', 'rules', 'Rules'),
    ('/search', 'search', 'Search'),
    ('/subscriptions/rules', 'subscriptionrules', 'Rules for Subscription'),
    ('/subscription', 'subscription', 'Subscription'),
    ('/subscriptions', 'subscriptions', 'Subscriptions'),
    ('/subscriptions_editor', 'subscriptions_editor', 'Subscriptions editor'),
    ('/suspicious_replicas', 'suspicious_replicas', 'Suspicious Replicas'),
    # TODO: add logfile extraction endpoints
    # ('/logfiles/load', 'load_logfile', load_logfile),
    # ('/logfiles/extract', 'extract_logfile', extract_logfile)
)

if POLICY == 'atlas':
    ATLAS_URLS = (
        ('/', 'atlas_index', 'Index'),
        ('/account_usage', 'account_usage', 'Group Account Usage'),
        ('/account_usage_history', 'account_usage_history', 'Account Usage History'),
        ('/dumps', 'dumps', 'Dumps'),
        ('/accounting', 'accounting', 'Accounting'),
        ('/conditions_summary', 'cond', 'ConditionsDB Summary'),
        ('/dbrelease_summary', 'dbrelease', 'DBRelease Summary'),
        ('/infrastructure', 'infrastructure', 'Infrastucture'),
        ('/rule_backlog_monitor', 'backlog_mon', 'Rules Backlog Monitoring')
    )
else:
    OTHER_URLS = (
        ('/', 'index', 'Index'),
    )


def view_maker(template, title):
    return lambda: authenticate(template=template, title=title)


def blueprint() -> Blueprint:
    bp = Blueprint('webui', __name__)

    for rule, endpoint, title in COMMON_URLS:
        template = endpoint + '.html'
        bp.add_url_rule(rule=rule, endpoint=endpoint, view_func=view_maker(template, title))

    for rule, endpoint, title in ATLAS_URLS:
        template = endpoint + '.html'
        bp.add_url_rule(rule=rule, endpoint=endpoint, view_func=view_maker(template, title))

    for rule, endpoint, title in OTHER_URLS:
        template = endpoint + '.html'
        bp.add_url_rule(rule=rule, endpoint=endpoint, view_func=view_maker(template, title))

    for rule, endpoint, view_func, methods in AUTH_URLS:
        bp.add_url_rule(rule=rule, endpoint=endpoint, view_func=view_func, methods=methods)

    return bp



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import subprocess
import sys
from typing import Union

clients_requirements_table = {
    'install_requires': [
        'click',
        'requests',
        'urllib3',
        'dogpile-cache',
        'packaging',
        'tabulate',
        'jsonschema',
        'rich',
        'typing_extensions'
    ],
    'extras_require': {
        'ssh': ['paramiko'],
        'kerberos': [
            'kerberos',
            'pykerberos',
            'requests-kerberos',
        ],
        'swift': ['python-swiftclient'],
        'argcomplete': ['argcomplete'],
        'sftp': ['paramiko'],
        'dumper': [
            'python-magic',
        ],
    }
}

dev_requirements = [
    'pytest',
    'pytest-xdist',
    'pytest-cov',
    'pyflakes',
    'flake8',
    'pylint',
    'isort',
    'xmltodict',
    'pytz',
    'pycodestyle',
    'pydoc-markdown',
    'docspec_python',
    'sh',
    'PyYAML',
]

server_requirements_table = {
    'install_requires': [
        'requests<=2.32.3',
        'urllib3<=1.26.19',
        'dogpile-cache<=1.2.2',
        'packaging<=24.1',
        'tabulate<=0.9.0',
        'jsonschema<=4.23.0',
        'rich<=13.9.4',
        'typing_extensions<=4.12.2',
        'argcomplete<=3.5.3',
        'boto',  # no upper limit is set in .in or .txt req files
        'python-magic<=0.4.27',
        'paramiko<=3.5.1',
        'boto3<=1.37.5',
        'sqlalchemy<=2.0.38',
        'alembic<=1.14.1',
        'pymemcache<=4.0.0',
        'python-dateutil<=2.9.0.post0',
        'stomp-py<=8.2.0',
        'statsd<=4.0.1',
        'geoip2<=5.0.1',
        'google-auth<=2.38.0',
        'redis<=5.2.1',
        'flask<=3.1.0',
        'oic<=1.7.0',
        'prometheus_client<=0.21.1',
    ],
    'extras_require': {
        'oracle': ['oracledb<=3.1.1'],
        'mongo': ['pymongo<=4.11.2'],
        'elastic': ['elasticsearch<=8.15.1'],
        'postgresql': [
            'psycopg<=3.2.3',
            'psycopg-binary<=3.2.3',
            'psycopg-pool<=3.2.3',
        ],
        'mysql': ['PyMySQL<=1.1.1'],
        'kerberos': [
            'kerberos<=1.3.1',
            'pykerberos<=1.2.4',
            'requests-kerberos<=0.15.0',
        ],
        'globus': [
            'PyYAML<=6.0.2',
            'globus-sdk<=3.41.0',
        ],
        'saml': ['python3-saml<=1.16.0'],
        'dev': dev_requirements
    }
}


def run_shell_command(cmd: str) -> Union[str, bytearray, memoryview]:
    """
    Run a shell command in path and return output"

    :param cmd: the shell command.
    :return: Output of the shell command.
    """
    output = subprocess.Popen(["/bin/sh", "-c", cmd], stdout=subprocess.PIPE)
    stdout = output.communicate()[0].strip()
    if isinstance(stdout, bytes):
        stdout = stdout.decode(errors='replace')
    return stdout


def get_rucio_version() -> str:
    python_executable = "'" + sys.executable + "'"
    ver = run_shell_command(
        "PYTHONPATH=lib " + python_executable + " -c "
        '"from rucio import version; print(version.version_string())"'
    )
    if not ver:
        raise RuntimeError("Could not fetch Rucio version")
    return str(ver)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging
import os
from unittest import mock

import pytest

import rucio.common.exception
import rucio.common.schema
import rucio.core.permission
from rucio.common.constants import DEFAULT_VO
from rucio.common.didtype import DID
from rucio.common.exception import DIDError
from rucio.common.plugins import PolicyPackageAlgorithms
from rucio.common.types import InternalAccount
from rucio.db.sqla.constants import DatabaseOperationType
from rucio.db.sqla.session import db_session


class TestPolicyPackageGeneric:

    @classmethod
    def setup_class(cls):
        # TODO: remove this after the generic schema becomes the default one used in tests (https://github.com/rucio/rucio/issues/7819)
        cls._rucio_default_vo_module = rucio.common.schema.schema_modules[DEFAULT_VO]
        rucio.common.schema.schema_modules[DEFAULT_VO] = importlib.import_module('rucio.common.schema.generic')

    @classmethod
    def teardown_class(cls):
        # TODO: remove this after the generic schema becomes the default one used in tests (https://github.com/rucio/rucio/issues/7819)
        rucio.common.schema.schema_modules[DEFAULT_VO] = cls._rucio_default_vo_module

    @pytest.mark.parametrize("name", [
        'name_with_underscore',
        'name-with-dash',
        'name.with.dot',
        'name/with/slash',
        '/name/starting/with/slash',
        'name/ending/with/slash/',
    ])
    def test_default_schema_did_name_valid(self, name, file_config_mock):
        rucio.common.schema.validate_schema('name', name, vo=DEFAULT_VO)

    @pytest.mark.parametrize("name", [
        '.startingWithDot',
        '-startingWithDash',
        '_startingWithUnderscore',
        '#hash',
        'name with spaces',
        'name*with*asterisk',
        'name\\with\\backslash',
        'name:with:colon',
        'name,with,comma',
    ])
    def test_default_schema_did_name_invalid(self, name, file_config_mock):
        with pytest.raises(rucio.common.exception.InvalidObject):
            rucio.common.schema.validate_schema('name', name, vo=DEFAULT_VO)


class TestPolicyPackage:

    @pytest.mark.noparallel(reason='changes global permission modules dictionary')
    # test that permission modules can implement a subset of actions and
    # fall back to generic for others
    def test_diff_based_permission(self):
        # replace permission module
        old_module = rucio.core.permission.permission_modules['def']
        rucio.core.permission.permission_modules['def'] = importlib.import_module('tests.mocks.permission_diff')

        # check that overridden action works as expected
        root_account = InternalAccount('root')

        with db_session(DatabaseOperationType.READ) as session:
            assert not rucio.core.permission.has_permission(root_account, 'add_account', {}, session=session)

            # check that omitted action falls back to generic module
            # root should be allowed to add RSE
            assert rucio.core.permission.has_permission(root_account, 'add_rse', {}, session=session)

        # restore original permission module
        rucio.core.permission.permission_modules['def'] = old_module

    @pytest.mark.noparallel(reason='changes global schema modules dictionary')
    # test that schema modules can define a subset of values and fall back
    # to generic for others
    def test_diff_based_schema(self):
        # replace schema module with our mock one
        old_module = rucio.common.schema.schema_modules['def']
        rucio.common.schema.schema_modules['def'] = importlib.import_module('tests.mocks.schema_diff')

        # check that overriden value is as expected
        assert rucio.common.schema.get_schema_value('SCOPE_LENGTH') == 50

        # check that omitted value falls back to generic module
        assert rucio.common.schema.get_schema_value('NAME_LENGTH') == 250

        # check that schemas defined in our module override generic
        rucio.common.schema.validate_schema('account', 'this-account-name-is-too-long-for-the-generic-schema-but-should-validate-against-the-mock-one')

        # check that schemas not defined in our module fall back to generic
        rucio.common.schema.validate_schema('r_name', 'name_to_validate')

        # restore original schema module
        rucio.common.schema.schema_modules['def'] = old_module

    @pytest.mark.noparallel(reason='changes environment variable')
    # test that a default algorithm will be loaded correctly
    def test_default_algorithm(self):
        # replace policy package in environment with our mock one
        old_pp_env = os.environ['RUCIO_POLICY_PACKAGE'] if 'RUCIO_POLICY_PACKAGE' in os.environ else None
        os.environ['RUCIO_POLICY_PACKAGE'] = 'tests.mocks.policy_package_algorithm'

        # retrieve default scope extraction algorithm
        algo = PolicyPackageAlgorithms._get_default_algorithm('scope')

        # call it to check we get the expected result
        assert algo('did') == 'Default scope algorithm loaded correctly!'

        # restore original policy package environment variable
        if old_pp_env is not None:
            os.environ['RUCIO_POLICY_PACKAGE'] = old_pp_env
        else:
            del os.environ['RUCIO_POLICY_PACKAGE']


@pytest.mark.noparallel(reason='Changes scope extraction policies')
@pytest.mark.parametrize("file_config_mock", [
        {"overrides": [("common", "extract_scope", "my_extraction_algo")]},
    ], indirect=True)
def test_change_scope_extraction(file_config_mock, did_factory, scope_factory, vo, rse_factory):

    rse, _ = rse_factory.make_posix_rse()
    mock_scope, _ = scope_factory(vos=[vo])

    def extract_scope(did, scopes, *args, **kwargs):
        return mock_scope, did.split(":")[-1]

    from rucio.common.utils import ScopeExtractionAlgorithms
    ScopeExtractionAlgorithms.register("my_extraction_algo", extract_scope)

    n_files = 3
    did_names = [did['did_name'] for did in did_factory.upload_test_dataset(rse_name=rse, scope=mock_scope, nb_files=n_files)]

    constucted_dids = [f"fake_scope:{did}" for did in did_names]

    for did in constucted_dids:
        derived_did = DID(did)
        assert derived_did.scope == mock_scope
        assert derived_did.name == did.split(':')[-1]


def test_scope_extraction_invalid_name(caplog):
    caplog.set_level(logging.DEBUG)  # Override log level
    with mock.patch('rucio.common.didtype.extract_scope', side_effect=ImportError()):

        DID("scope:name")
        assert "Failure using extract_scope policy for 'scope:name'" in caplog.text

        with pytest.raises(DIDError):
            DID("invalid/scope/name/combo")



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import choice
from string import ascii_lowercase, ascii_uppercase

import pytest

from rucio.common.exception import InvalidRSEExpression, RSEWriteBlocked
from rucio.core import rse, rse_expression_parser


def attribute_name_generator(size=10):
    return ''.join(choice(ascii_uppercase)).join(choice(ascii_lowercase) for x in range(size - 1))


@pytest.mark.noparallel(reason='uses pre-defined RSE, test_all_rse fails when run in parallel')
class TestRSEExpressionParserCore:

    @classmethod
    @pytest.fixture(scope='class')
    def setup_test_env(cls, rse_factory_unittest, tag_factory_class):
        already_existing_rses = rse.list_rses()

        rse1, rse1_id = rse_factory_unittest.make_mock_rse()
        rse2, rse2_id = rse_factory_unittest.make_mock_rse()
        rse3, rse3_id = rse_factory_unittest.make_mock_rse()
        rse4, rse4_id = rse_factory_unittest.make_mock_rse()
        rse5, rse5_id = rse_factory_unittest.make_mock_rse()

        # Add Attributes
        attribute = attribute_name_generator()

        rse.add_rse_attribute(rse1_id, attribute, "at")
        rse.add_rse_attribute(rse2_id, attribute, "de")
        rse.add_rse_attribute(rse3_id, attribute, "fr")
        rse.add_rse_attribute(rse4_id, attribute, "uk")
        rse.add_rse_attribute(rse5_id, attribute, "us")

        # Add numeric Attributes
        attribute_numeric = attribute_name_generator()

        rse.add_rse_attribute(rse1_id, attribute_numeric, 10)
        rse.add_rse_attribute(rse2_id, attribute_numeric, 20)
        rse.add_rse_attribute(rse3_id, attribute_numeric, 30)
        rse.add_rse_attribute(rse4_id, attribute_numeric, 40)
        rse.add_rse_attribute(rse5_id, attribute_numeric, 50)

        # Add Tags
        tag1 = tag_factory_class.new_tag()
        tag2 = tag_factory_class.new_tag()
        rse.add_rse_attribute(rse1_id, tag1, True)
        rse.add_rse_attribute(rse2_id, tag1, True)
        rse.add_rse_attribute(rse3_id, tag1, True)
        rse.add_rse_attribute(rse4_id, tag2, True)
        rse.add_rse_attribute(rse5_id, tag2, True)

        return (rse1, rse1_id, rse2, rse2_id, rse3, rse3_id, rse4, rse4_id, rse5, rse5_id,
                attribute, attribute_numeric, tag1, tag2, already_existing_rses)

    @pytest.fixture(autouse=True)
    def setup_obj(self, setup_test_env, vo):
        self.filter = {'filter_': {'vo': vo}}
        (
            self.rse1, self.rse1_id, self.rse2, self.rse2_id, self.rse3, self.rse3_id, self.rse4, self.rse4_id, self.rse5, self.rse5_id,
            self.attribute, self.attribute_numeric, self.tag1, self.tag2, self.already_existing_rses,
        ) = setup_test_env

    def test_unconnected_operator(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test invalid rse expression: unconnected operator"""
        with pytest.raises(InvalidRSEExpression):
            rse_expression_parser.parse_expression("TEST_RSE1|", **self.filter)

    def test_wrong_parantheses(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test invalid rse expression: wrong parentheses """
        with pytest.raises(InvalidRSEExpression):
            rse_expression_parser.parse_expression("TEST_RSE1)", **self.filter)

    def test_unknown_rse(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test unknown RSE """
        with pytest.raises(InvalidRSEExpression):
            rse_expression_parser.parse_expression("TEST_RSE999", **self.filter)

    def test_simple_rse_reference(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test simple RSE reference """
        value = [t_rse['id'] for t_rse in rse_expression_parser.parse_expression(self.rse1, **self.filter)]
        assert value == [self.rse1_id]

    def test_attribute_reference(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test simple RSE attribute reference """
        value = [t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s=uk" % self.attribute, **self.filter)]
        assert value == [self.rse4_id]

    def test_all_rse(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test reference on all RSE """
        all_rses = rse.list_rses(filters=self.filter['filter_'])
        rse_expression_parser.REGION.invalidate()
        value = rse_expression_parser.parse_expression("*", **self.filter)
        for rse_ in self.already_existing_rses:
            if rse_ in all_rses:
                all_rses.remove(rse_)
            if rse_ in value:
                value.remove(rse_)
        value = sorted(value, key=lambda rse_: rse_['rse'])
        expected = sorted(all_rses, key=lambda rse_: rse_['rse'])
        assert len(value) == 5
        assert value == expected

    def test_tag_reference(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test simple RSE tag reference """
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression(self.tag1, **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id, self.rse3_id])
        assert value == expected

    def test_parantheses(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test parentheses """
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("(%s)" % self.tag1, **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id, self.rse3_id])
        assert value == expected

    def test_union(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test union operator """
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s|%s" % (self.tag1, self.tag2), **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id, self.rse3_id, self.rse4_id, self.rse5_id])
        assert value == expected

    def test_complement(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test complement operator """
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s\\%s" % (self.tag1, self.rse3), **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id])
        assert value == expected

    def test_intersect(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test intersect operator """
        value = [t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s&%s=uk" % (self.tag2, self.attribute), **self.filter)]
        assert value == [self.rse4_id]

    def test_order_of_operations(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test order of operations """
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s\\%s|%s=fr" % (self.tag1, self.rse3, self.attribute), **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id, self.rse3_id])
        assert value == expected
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s\\(%s|%s=fr)" % (self.tag1, self.rse3, self.attribute), **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id])
        assert value == expected

    def test_complicated_expression_1(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test some complicated expression 1"""
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("(%s|%s)\\%s|%s&%s" % (self.tag1, self.tag2, self.tag2, self.tag2, self.tag1), **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id, self.rse3_id])
        assert value == expected

    def test_complicated_expression_2(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test some complicated expression 2"""
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("(((((%s))))|%s=us)&%s|(%s=at|%s=de)" % (self.tag1, self.attribute, self.tag2, self.attribute, self.attribute), **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id, self.rse5_id])
        assert value == expected

    def test_complicated_expression_3(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test some complicated expression 3"""
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("(*)&%s=at" % self.attribute, **self.filter)])
        expected = sorted([self.rse1_id])
        assert value == expected

    def test_list_on_availability(self, rse_factory):
        """ RSE_EXPRESSION_PARSER (CORE) List rses based on availability filter"""
        rsewrite_name, rsewrite_id = rse_factory.make_mock_rse()
        rsenowrite_name, rsenowrite_id = rse_factory.make_mock_rse()

        attribute = attribute_name_generator()

        rse.add_rse_attribute(rsewrite_id, attribute, "de")
        rse.add_rse_attribute(rsenowrite_id, attribute, "de")

        rse.update_rse(rsewrite_id, {'availability_write': True})
        rse.update_rse(rsenowrite_id, {'availability_write': False})

        value = sorted([item['id'] for item in rse_expression_parser.parse_expression("%s=de" % attribute, **self.filter)])
        expected = sorted([rsewrite_id, rsenowrite_id])
        assert value == expected

        filters = self.filter
        filters['availability_write'] = True
        value = sorted([item['id'] for item in rse_expression_parser.parse_expression("%s=de" % attribute, filters)])
        expected = sorted([rsewrite_id])
        assert value == expected

        filters['availability_write'] = False
        pytest.raises(RSEWriteBlocked, rse_expression_parser.parse_expression, "%s=de" % attribute, filters)

    def test_numeric_operators(self):
        """ RSE_EXPRESSION_PARSER (CORE) Test RSE attributes with numeric operations """
        value = [t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s<11" % self.attribute_numeric, **self.filter)]
        assert value == [self.rse1_id]
        pytest.raises(InvalidRSEExpression, rse_expression_parser.parse_expression, "%s<9" % self.attribute_numeric, **self.filter)
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s<21" % self.attribute_numeric, **self.filter)])
        expected = sorted([self.rse1_id, self.rse2_id])
        assert value == expected
        value = [t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s>49" % self.attribute_numeric, **self.filter)]
        assert value == [self.rse5_id]
        pytest.raises(InvalidRSEExpression, rse_expression_parser.parse_expression, "%s>51" % self.attribute_numeric, **self.filter)
        value = sorted([t_rse['id'] for t_rse in rse_expression_parser.parse_expression("%s>30" % self.attribute_numeric, **self.filter)])
        expected = sorted([self.rse4_id, self.rse5_id])
        assert value == expected


@pytest.mark.noparallel(reason='uses pre-defined RSE')
class TestRSEExpressionParserClient:

    @pytest.fixture(autouse=True)
    def setup_obj(self, rse_factory, tag_factory):
        self.rse1, self.rse1_id = rse_factory.make_mock_rse()
        self.rse2, self.rse2_id = rse_factory.make_mock_rse()
        self.rse3, self.rse3_id = rse_factory.make_mock_rse()
        self.rse4, self.rse4_id = rse_factory.make_mock_rse()
        self.rse5, self.rse5_id = rse_factory.make_mock_rse()

        # Add Attributes
        self.attribute = attribute_name_generator()

        rse.add_rse_attribute(self.rse1_id, self.attribute, "at")
        rse.add_rse_attribute(self.rse2_id, self.attribute, "de")
        rse.add_rse_attribute(self.rse3_id, self.attribute, "fr")
        rse.add_rse_attribute(self.rse4_id, self.attribute, "uk")
        rse.add_rse_attribute(self.rse5_id, self.attribute, "us")

        # Add Tags
        self.tag1 = tag_factory.new_tag()
        self.tag2 = tag_factory.new_tag()
        rse.add_rse_attribute(self.rse1_id, self.tag1, True)
        rse.add_rse_attribute(self.rse2_id, self.tag1, True)
        rse.add_rse_attribute(self.rse3_id, self.tag1, True)
        rse.add_rse_attribute(self.rse4_id, self.tag2, True)
        rse.add_rse_attribute(self.rse5_id, self.tag2, True)

    def test_complicated_expression(self, rse_client):
        """ RSE_EXPRESSION_PARSER (CLIENT) Test some complicated expression"""
        rses = sorted([item['rse'] for item in rse_client.list_rses("(((((%s))))|%s=us)&%s|(%s=at|%s=de)" % (self.tag1, self.attribute, self.tag2, self.attribute, self.attribute))])
        expected = sorted([self.rse1, self.rse2, self.rse5])
        assert rses == expected

    def test_complicated_expression_1(self, rse_client):
        """ RSE_EXPRESSION_PARSER (CORE) Test some complicated expression 1"""
        rses = sorted([item['rse'] for item in rse_client.list_rses("(%s|%s)\\%s|%s&%s" % (self.tag1, self.tag2, self.tag2, self.tag2, self.tag1))])
        expected = sorted([self.rse1, self.rse2, self.rse3])
        assert rses == expected



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta

import pytest

from rucio.common.constants import HTTPMethod
from rucio.common.exception import CannotAuthenticate, ClientProtocolNotFound, ClientProtocolNotSupported, MissingClientParameter, RucioException
from rucio.common.utils import execute
from rucio.tests.common import remove_config, skip_outside_gh_actions
from tests.mocks.mock_http_server import MockServer


@pytest.fixture
def client_token_path_override(file_config_mock, function_scope_prefix, tmp_path):
    """
    Ensure each running client has a different path for the token, otherwise tests cannot run in parallel
    """
    from rucio.common.config import config_set
    config_set('client', 'auth_token_file_path', str(tmp_path / f'{function_scope_prefix}token'))


@pytest.mark.usefixtures("client_token_path_override")
class TestBaseClient:
    """ To test Clients"""

    from rucio.common.config import config_get

    cacert = config_get('test', 'cacert')
    usercert = config_get('test', 'usercert')
    userkey = config_get('test', 'userkey')

    def test_userpass(self, vo):
        """ CLIENTS (BASECLIENT): authenticate with userpass."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient
        client = BaseClient(account='root', ca_cert=self.cacert, auth_type='userpass', creds=creds, vo=vo)
        print(client)

    def test_userpass_wrong_creds(self, vo):
        """ CLIENTS (BASECLIENT): try to authenticate with wrong username."""
        creds = {'username': 'wrong', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(CannotAuthenticate):
            BaseClient(account='root', ca_cert=self.cacert, auth_type='userpass', creds=creds, vo=vo)

    def test_userpass_no_ca_cert(self, vo):
        """ CLIENTS (BASECLIENT): authenticate with userpass without ca cert."""
        creds = {'username': 'wrong', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(CannotAuthenticate):
            BaseClient(account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_userpass_no_creds_passed(self, vo):
        """ CLIENTS (BASECLIENT): authenticate with userpass, with no credentials passed"""
        from rucio.client.baseclient import BaseClient

        client = BaseClient(account='root', ca_cert=self.cacert, auth_type='userpass', creds=None, vo=vo)
        assert client.creds['username'] == 'ddmlab'
        assert client.creds['password'] == 'secret'

    @skip_outside_gh_actions
    def test_x509(self, vo):
        """ CLIENTS (BASECLIENT): authenticate with x509."""
        from rucio.client.baseclient import BaseClient

        creds = {'client_cert': self.usercert,
                 'client_key': self.userkey}
        BaseClient(account='root', ca_cert=self.cacert, auth_type='x509', creds=creds, vo=vo)

    @pytest.mark.parametrize("cert_path", ["", '/opt/rucio/etc/web/notthere.crt'])
    def test_x509_non_existing_cert(self, vo, cert_path):
        """ CLIENTS (BASECLIENT): authenticate with x509 with missing certificate."""
        creds = {'client_cert': cert_path}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(MissingClientParameter, match=rf"X\.509 client certificate not found: '{cert_path}'"):
            BaseClient(account='root', ca_cert=self.cacert, auth_type='x509', creds=creds, vo=vo)

    def test_client_rucio_protocol_not_supported(self, vo):
        """ CLIENTS (BASECLIENT): try to pass a rucio host with a not supported protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(ClientProtocolNotSupported):
            BaseClient(rucio_host='junk://localhost', auth_host='http://localhost', account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_client_auth_protocol_not_supported(self, vo):
        """ CLIENTS (BASECLIENT): try to pass an auth host with a not supported protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(ClientProtocolNotSupported):
            BaseClient(rucio_host='https://localhost', auth_host='junk://localhost', account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_client_rucio_and_auth_protocol_not_supported(self, vo):
        """ CLIENTS (BASECLIENT): try to pass both a rucio and auth host with a not supported protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(ClientProtocolNotSupported):
            BaseClient(rucio_host='junk://localhost', auth_host='junk://localhost', account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_client_rucio_protocol_not_found(self, vo):
        """ CLIENTS (BASECLIENT): try to pass a rucio host with a missing protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(ClientProtocolNotFound):
            BaseClient(rucio_host='localhost', auth_host='https://localhost', account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_client_auth_protocol_not_found(self, vo):
        """ CLIENTS (BASECLIENT): try to pass an auth host with a not supported protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(ClientProtocolNotFound):
            BaseClient(rucio_host='http://localhost', auth_host='localhost', account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_client_rucio_and_auth_protocol_not_found(self, vo):
        """ CLIENTS (BASECLIENT): try to pass both a rucio and auth host with a missing protocol."""
        creds = {'username': 'ddmlab', 'password': 'secret'}
        from rucio.client.baseclient import BaseClient

        with pytest.raises(ClientProtocolNotFound):
            BaseClient(rucio_host='localhost', auth_host='localhost', account='root', auth_type='userpass', creds=creds, vo=vo)

    def test_retry_on_502_always_fail(self, vo):
        """ CLIENTS (BASECLIENT): Ensure client retries on 502 error codes, but fails on repeated errors"""
        from rucio.client.baseclient import BaseClient

        class AlwaysFailWith502(MockServer.Handler):
            def do_GET(self):
                self.send_code_and_message(502, {}, '')

        with MockServer(AlwaysFailWith502) as server:
            with pytest.raises(CannotAuthenticate):
                creds = {'username': 'ddmlab', 'password': 'secret'}
                BaseClient(rucio_host=server.base_url, auth_host=server.base_url, account='root', auth_type='userpass', creds=creds, vo=vo)

            with pytest.raises(RucioException):
                creds = {'client_cert': self.usercert,
                         'client_key': self.userkey}
                BaseClient(rucio_host=server.base_url, auth_host=server.base_url, account='root', ca_cert=self.cacert, auth_type='x509', creds=creds, vo=vo)

    def test_retry_on_502_succeeds_eventually(self, vo):
        """ CLIENTS (BASECLIENT): Ensure client retries on 502 error codes"""
        invocations = []
        from rucio.client.baseclient import BaseClient

        class FailTwiceWith502(MockServer.Handler):
            def do_GET(self, invocations=invocations):
                invocations.append(self.path)
                if len(invocations) <= 2:
                    self.send_code_and_message(502, {}, '')
                else:
                    self.send_code_and_message(200, {'x-rucio-auth-token': 'sometoken'}, '')

        start_time = datetime.utcnow()
        with MockServer(FailTwiceWith502) as server:
            creds = {'username': 'ddmlab', 'password': 'secret'}
            del invocations[:]
            client = BaseClient(rucio_host=server.base_url, auth_host=server.base_url, account='root', auth_type='userpass', creds=creds, vo=vo)
            del invocations[:]
            client._send_request(server.base_url, method=HTTPMethod.GET)  # noqa
        # The client did back-off multiple times before succeeding: 2 * 0.25s (authentication) + 2 * 0.25s (request) = 1s
        assert datetime.utcnow() - start_time > timedelta(seconds=0.9)


class TestRucioClients:
    """ To test Clients"""

    marker = '$> '

    def test_ping(self, vo):
        """ PING (CLIENT): Ping Rucio """

        creds = {'username': 'ddmlab', 'password': 'secret'}

        from rucio.client.client import Client
        from rucio.common.config import config_get

        cacert = config_get('test', 'cacert')
        client = Client(account='root', ca_cert=cacert, auth_type='userpass', creds=creds, vo=vo)
        print(client.ping())

    @pytest.mark.noparallel(reason='We temporarily remove the config file.')
    @remove_config
    def test_import_without_config_file(self, vo):
        """
        The Client should be importable without a config file, since it is
        configurable afterwards.

        We are in a fully configured environment with a default config file. We
        thus have to disable the access to it (move it) and make sure to run the
        code in a different environment.
        """

        exitcode, _, err = execute("python -c 'from rucio.client import Client'")
        print(exitcode, err)
        assert exitcode == 0
        assert "Could not load Rucio configuration file." not in err

    @pytest.mark.noparallel(reason='We temporarily remove the config file.')
    @remove_config
    def test_run_client_no_config(self, vo):
        rucio_host = "https://rucio:443"
        auth_host = "https://rucio:443"
        ca_cert = "/etc/grid-security/certificates/5fca1cb1.0"
        creds = {'username': 'ddmlab', 'password': 'secret'}

        from rucio.client.client import Client

        client = Client(
            rucio_host=rucio_host,
            auth_host=auth_host,
            ca_cert=ca_cert,
            account='root',
            auth_type='userpass',
            creds=creds,
            vo=vo)

        # Couple of basic calls
        # account
        assert client.whoami()['account'] == 'root'
        assert client.get_account('jdoe')['account'] == "jdoe"

        # config
        assert client.set_config_option("mock_section", "mock_option", value=0)

        # DID
        assert list(client.list_dids(scope='mock', filters={})) is not None

        # replica
        assert list(client.list_replicas(dids=[])) is not None

        # rse
        assert list(client.list_rses()) is not None

        # rule
        assert list(client.list_replication_rules()) is not None



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from rucio.common.exception import InsufficientAccountLimit, InsufficientTargetRSEs
from rucio.core.account_counter import increase, update_account_counter
from rucio.core.account_limit import set_global_account_limit, set_local_account_limit
from rucio.core.rse_selector import RSESelector
from rucio.db.sqla.constants import DatabaseOperationType
from rucio.db.sqla.session import db_session


@pytest.fixture
def test_rses(rse_factory):
    rse1_name, rse1_id = rse_factory.make_mock_rse()
    rse2_name, rse2_id = rse_factory.make_mock_rse()

    rse1 = {'id': rse1_id, 'staging_area': False}
    rse2 = {'id': rse2_id, 'staging_area': False}

    return rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2


class TestRSESelectorInit:

    def test_1(self, random_account, test_rses):
        # more copies than RSEs -> error
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        rses = [rse1]
        copies = 2
        with pytest.raises(InsufficientTargetRSEs):
            RSESelector(random_account, rses, None, copies)

    def test_2(self, random_account, test_rses):
        # local quota not enough -> error
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        copies = 2
        rses = [rse1, rse2]
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=10)
        with db_session(DatabaseOperationType.WRITE) as session:
            increase(rse1_id, random_account, 10, 10, session=session)
            update_account_counter(account=random_account, rse_id=rse1_id, session=session)
        with pytest.raises(InsufficientAccountLimit):
            RSESelector(random_account, rses, None, copies)

    def test_3(self, random_account, test_rses):
        # global quota not enough -> error
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        copies = 2
        rses = [rse1, rse2]
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=20)
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=10)
        with db_session(DatabaseOperationType.WRITE) as session:
            increase(rse1_id, random_account, 10, 10, session=session)
            update_account_counter(account=random_account, rse_id=rse1_id, session=session)
        with pytest.raises(InsufficientAccountLimit):
            RSESelector(random_account, rses, None, copies)

    def test_4(self, random_account, test_rses):
        # enough RSEs, local and global quota -> 2 RSEs
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=20)
        set_global_account_limit(account=random_account, rse_expression=rse2_name, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse2_id, bytes_=20)
        copies = 2
        rses = [rse1, rse2]
        rse_selector = RSESelector(random_account, rses, None, copies)
        assert len(rse_selector.rses) == 2

    def test_5(self, random_account, test_rses):
        # enough RSEs and local quota, but global quota missing -> 1 RSE
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        copies = 1
        rses = [rse1, rse2]
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=10)
        with db_session(DatabaseOperationType.WRITE) as session:
            increase(rse1_id, random_account, 10, 10, session=session)
            update_account_counter(account=random_account, rse_id=rse1_id, session=session)
        set_local_account_limit(account=random_account, rse_id=rse2_id, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=20)
        rse_selector = RSESelector(random_account, rses, None, copies)
        assert len(rse_selector.rses) == 1

    def test_6(self, random_account, test_rses):
        # enough RSEs and global quota, but local quota missing -> 1 RSE
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        rses = [rse1, rse2]
        copies = 1
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=10)
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=10)
        with db_session(DatabaseOperationType.WRITE) as session:
            increase(rse1_id, random_account, 10, 10, session=session)
            update_account_counter(account=random_account, rse_id=rse1_id, session=session)
        set_local_account_limit(account=random_account, rse_id=rse2_id, bytes_=10)
        set_global_account_limit(account=random_account, rse_expression=rse2_name, bytes_=10)
        rse_selector = RSESelector(random_account, rses, None, copies)
        assert len(rse_selector.rses) == 1


class TestRSESelectorDynamic:

    def test_1(self, random_account, test_rses):
        # enough RSEs and global quota, but not enough local quota after change -> 1 RSE
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=20)
        set_global_account_limit(account=random_account, rse_expression=rse2_name, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=10)
        set_local_account_limit(account=random_account, rse_id=rse2_id, bytes_=10)
        copies = 2
        rses = [rse1, rse2]
        rse_selector = RSESelector(random_account, rses, None, copies)
        assert len(rse_selector.rses) == 2
        rse_selector.select_rse(9, [rse1_id], copies=1)
        rses = rse_selector.select_rse(5, [], copies=1)
        assert len(rses) == 1
        assert rses[0][0] == rse2_id

    def test_2(self, random_account, test_rses):
        # enough RSEs and global quota, but not enough global quota after change -> 1 RSE
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=10)
        set_global_account_limit(account=random_account, rse_expression=rse2_name, bytes_=10)
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse2_id, bytes_=20)
        copies = 2
        rses = [rse1, rse2]
        rse_selector = RSESelector(random_account, rses, None, copies)
        assert len(rse_selector.rses) == 2
        rse_selector.select_rse(10, [rse1_id], copies=1)
        rses = rse_selector.select_rse(5, [], copies=1)
        assert len(rses) == 1
        assert rses[0][0] == rse2_id

    def test_3(self, random_account, test_rses):
        # enough RSEs and global quota, also after after change -> 2 RSE
        rse1_name, rse1_id, rse1, rse2_name, rse2_id, rse2 = test_rses
        set_global_account_limit(account=random_account, rse_expression=rse1_name, bytes_=20)
        set_global_account_limit(account=random_account, rse_expression=rse2_name, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=20)
        set_local_account_limit(account=random_account, rse_id=rse2_id, bytes_=20)
        copies = 2
        rses = [rse1, rse2]
        rse_selector = RSESelector(random_account, rses, None, copies)
        assert len(rse_selector.rses) == 2
        rse_selector.select_rse(10, [rse1_id], copies=1)
        rse_selector.select_rse(10, [rse2_id], copies=1)
        rses = rse_selector.select_rse(5, [], copies=2)
        assert len(rses) == 2
